#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements functions to initiate a DataStore without uploading any data.
"""
from typing import List, Union, Generator
import io

import pandas as pd

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.builder.builder_resource import builder_request_default_auth_if_ckan, initial_resource_building_state
from ckanapi_harvesters.builder.builder_resource_datastore_file import BuilderDataStoreFile
from ckanapi_harvesters.builder.builder_resource_multi_abc import FileChunkDataFrame
from ckanapi_harvesters.auxiliary.ckan_errors import NotMappedObjectNameError, DataStoreNotFoundError
from ckanapi_harvesters.auxiliary.list_records import ListRecords
from ckanapi_harvesters.builder.builder_errors import ResourceFileNotExistMessage
from ckanapi_harvesters.auxiliary.ckan_model import CkanResourceInfo
from ckanapi_harvesters.auxiliary.ckan_errors import CkanArgumentError, FunctionMissingArgumentError
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element, assert_or_raise, datastore_id_col


class BuilderDataStoreUrl(BuilderDataStoreFile):  #, BuilderUrlABC):  # multiple inheritance can give undefined results
    """
    Class representing a DataStore (resource metadata and fields metadata) defined by a url.
    """
    def __init__(self, *, parent, name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None, url:str=None, options_string:str=None, base_dir:str=None):
        super(BuilderDataStoreFile, self).__init__(parent=parent, name=name, format=format, description=description, resource_id=resource_id,
                                                   download_url=download_url, options_string=options_string, base_dir=base_dir)
        # super(BuilderUrlABC, self).__init__(name=name, format=format, description=description, resource_id=resource_id, download_url=download_url, url=url)
        self.reupload_on_update = False
        self.reupload_if_needed = False
        self.url:str = url
        self.file_name = name

    def copy(self, *, dest=None, parent=None):
        if dest is None:
            dest = BuilderDataStoreUrl(parent=self.parent_package)
        super().copy(dest=dest, parent=parent)
        dest.reupload_on_update = self.reupload_on_update
        dest.reupload_if_needed = self.reupload_if_needed
        dest.url = self.url
        dest.file_name = self.file_name
        return dest

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super(BuilderDataStoreFile, self)._load_from_df_row(row=row)
        # super(BuilderUrlABC, self)._load_from_df_row(row=row)
        self.url: str = _string_from_element(row["file/url"], strip=True)
        self.file_name = self.name

    @staticmethod
    def sample_file_path_is_url() -> bool:
        return True

    def get_sample_file_path(self, resources_base_dir: str, ckan:Union[CkanApi,None]=None, file_index:int=0) -> str:
        return self.url

    def init_local_files_list(self, resources_base_dir:str, cancel_if_present:bool=True, **kwargs) -> List[str]:
        file_path = self.get_sample_file_path(resources_base_dir=resources_base_dir)
        self.file_size = 1
        self.local_file_list = [file_path]
        self.local_file_size = [self.file_size]
        self.local_file_size_sum = 1
        return self.local_file_list

    def load_sample_data(self, resources_base_dir:str, *, ckan:CkanApi=None,
                         proxies:dict=None, headers:dict=None) -> bytes:
        self.sample_source = self.url
        if ckan is None:
            raise FunctionMissingArgumentError("BuilderDataStoreUrl.load_sample_data", "ckan")
        return ckan.download_url_proxy(self.url, proxies=proxies, headers=headers, auth_if_ckan=builder_request_default_auth_if_ckan).content

    def get_local_df_chunk_generator(self, resources_base_dir:str, ckan:CkanApi, **kwargs) -> Generator[FileChunkDataFrame, None, None]:
        payload = self.load_sample_data(resources_base_dir=resources_base_dir)
        with io.StringIO(payload.decode()) as stream:
            self.read_line_counter = 0
            with self.local_file_format.read_buffer_full(stream, fields=self._get_fields_info()) as df_file:
                if isinstance(df_file, pd.DataFrame) or isinstance(df_file, ListRecords):
                    self.file_semaphore.acquire()
                    self.read_line_counter += len(df_file)
                    line_counter = self.read_line_counter
                    self.file_semaphore.release()
                    yield FileChunkDataFrame(df_file, self.url, 0, 0, 0, line_counter)
                else:  # iterator
                    with df_file as df_generator:
                        if hasattr(df_file, "handles"):
                            file_handle = df_file.handles.handle
                        else:
                            file_handle = None
                        previous_file_position = 0
                        # for chunk_index, df in enumerate(df_generator):
                        chunk_index = 0
                        file_position = 0
                        while True:
                            self.file_semaphore.acquire()
                            if file_handle is not None:
                                file_position = file_handle.buffer.tell()  # approximative position in file
                            else:
                                file_position = 0  # no position tracking available
                            try:
                                df = next(df_generator)
                                if file_handle is None and hasattr(df, "attrs") and "file_position" in df.attrs:
                                    file_position = df.attrs["file_position"]
                                self.read_line_counter += len(df)
                                line_counter = self.read_line_counter
                            except StopIteration:
                                self.file_semaphore.release()
                                break
                            self.file_semaphore.release()
                            yield FileChunkDataFrame(df, self.url, 0, chunk_index, previous_file_position, line_counter)
                            previous_file_position = file_position
                            chunk_index = chunk_index + 1

    def upload_request_full(self, ckan:CkanApi, resources_base_dir:str, *,
                            threads:int=1, external_stop_event=None,
                            start_index:int=0, end_index:int=None, **kwargs) -> None:
        # CKAN manages URL resources: no upload is necessary
        return

    @staticmethod
    def resource_mode_str() -> str:
        return "DataStore from URL"

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = self.url
        return d

    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[None,ContextErrorLevelMessage]:
        if ckan is None:
            return ResourceFileNotExistMessage(self.name, ErrorLevel.Warning, "Could not determine if resource url exists because ckan argument was not provided.")
        else:
            return ckan.download_url_proxy_test_head(self.url, **kwargs)

    def patch_request(self, ckan: CkanApi, package_id: str, *,
                      df_upload:pd.DataFrame=None, payload:Union[bytes, io.BufferedIOBase]=None,
                      reupload: bool = None, override_ckan:bool=False, resources_base_dir:str=None) -> CkanResourceInfo:
        """
        Specific implementation of patch_request which does not upload any data and only updates the fields currently present in the database

        :param resources_base_dir:
        :param ckan:
        :param package_id:
        :param reupload:
        :return:
        """
        self._merge_resource_attributes(override_ckan=override_ckan)
        if reupload is None: reupload = self.reupload_on_update
        if payload is not None or df_upload is not None:
            raise CkanArgumentError("payload", "datastore defined from URL patch")
        resource_id = self.get_or_query_resource_id(ckan, error_not_found=False)
        try:
            df_download = self.download_resource_df(ckan, download_alter=False, search_all=False, limit=1)
            if df_download is None:
                assert_or_raise(resource_id is None, RuntimeError("Unexpected: resource_id should be None"))
                raise NotMappedObjectNameError(self.name)
            current_df_fields = set(df_download.columns)
        except NotMappedObjectNameError as e:
            df_download = None
            current_df_fields = set()
        except DataStoreNotFoundError as e:
            df_download = None
            current_df_fields = set()
        empty_datastore = df_download is None or len(df_download) == 0
        data_cleaner_fields = None
        data_cleaner_index = set()
        current_df_fields -= {datastore_id_col}  # _id does not require documentation
        aliases = self._get_alias_list(ckan)
        self._check_necessary_fields(current_df_fields, raise_error=False, empty_datastore=empty_datastore)
        self._check_undocumented_fields(current_df_fields)
        primary_key, indexes = self._get_primary_key_indexes(data_cleaner_index, current_df_fields=current_df_fields,
                                                             error_missing=False, empty_datastore=empty_datastore)
        fields_update = self._get_fields_update(ckan, current_df_fields=current_df_fields, data_cleaner_fields=data_cleaner_fields,
                                                reupload=reupload, override_ckan=override_ckan)
        fields = list(fields_update.values()) if len(fields_update) > 0 else None
        resource_info = ckan.resource_create(package_id, name=self.name, format=self.resource_attributes.format, description=self.resource_attributes.description,
                                             state=initial_resource_building_state if initial_resource_building_state is not None else self.resource_attributes.state,
                                             url=self.url,
                                             datastore_create=False, auto_submit=False, create_default_view=self.create_default_view,
                                             cancel_if_exists=True, update_if_exists=True, aliases=aliases, reupload=False, data_cleaner=self.data_cleaner_upload,
                                             progress_callback=self.progress_callback)
        resource_id = resource_info.id
        self.known_id = resource_id
        self._compare_fields_to_datastore_info(resource_info, current_df_fields, ckan)
        if reupload:
            # re-initialize datastore to reupload from url
            # normally, data was automatically submitted to DataStore on resource_create (not needed)
            ckan.datastore_create(resource_id, fields=fields, primary_key=primary_key, indexes=indexes, aliases=aliases)
            ckan.datastore_submit(resource_id)
        return resource_info






#!python3
# -*- coding: utf-8 -*-
"""
Data format policy representation and enforcing for lists of tags grouped in vocabularies
"""
from typing import List, Dict

from ckanapi_harvesters.policies.data_format_policy_lists import ValueListPolicy, GroupedValueListPolicy

tag_subs_re = r"[^a-zA-Z0-9_\-\.]"


class TagListPolicy(ValueListPolicy):
    def get_tags_list_dict(self, vocabulary_id: str=None) -> List[Dict[str, str]]:
        """
        Generate tags dictionary to initiate a vocabulary using the CKAN API.

        :param vocabulary_id:
        :return:
        """
        if vocabulary_id is not None:
            tags_list_dict = [{"name": tag_spec.value, "vocabulary_id": vocabulary_id} for tag_spec in self.list_specs]
        else:
            tags_list_dict = [{"name": tag_spec.value} for tag_spec in self.list_specs]
        return tags_list_dict


class TagGroupsListPolicy(GroupedValueListPolicy):
    pass





#!python3
# -*- coding: utf-8 -*-
"""
Upload a sample dataset from the example package (needs to be online)

Samples are used to expose a portion of a dataset to a user, which can be publicly exposed,
so he can contact the data provider to demand full access.
"""
from ckanapi_harvesters.builder.builder_package import BuilderPackage
from ckanapi_harvesters.ckan_api import CkanApi

from ckanapi_harvesters.builder.example import example_package_xls


def run(ckan:CkanApi = None):
    BuilderPackage.unlock_external_code_execution()

    mdl_xls = BuilderPackage.from_excel(example_package_xls)
    ckan = mdl_xls.init_ckan(ckan)
    ckan.input_missing_info(input_args_if_necessary=True, input_owner_org=True)
    ckan.set_limits(10000)  # reduce if server hangs up
    ckan.set_submit_timeout(5)
    ckan.set_verbosity(True)
    ckan.test_ckan_login(raise_error=True, verbose=True)
    # ckan.set_default_map_mode(datastore_info=True)  # uncomment to query DataStore information

    mdl_ckan = BuilderPackage.from_ckan(ckan, mdl_xls.package_name)

    sample_mdl = mdl_ckan.setup_sample_package(ckan)
    sample_df_dict = mdl_ckan.download_sample(ckan, limit=10)  # download the 10 first lines of each resource
    # sample_df_dict = mdl_ckan.download_sample(ckan, limit=10, include_files=False, empty_files=True)  # these options empty the file resources

    sample_df_dict["traces.csv"] = mdl_ckan.resource_builders["traces.csv"].download_sample_df(ckan, limit=120)  # specific request for this resource: take 120 lines
    # sample_df_dict["users.csv"] = pd.DataFrame()  # erase DataStore contents
    # sample_df_dict["generate_example.py"] = bytes()  # erase file contents

    sample_mdl.patch_request_full(ckan, reupload=True, sample_df_dict=sample_df_dict)
    sample_mdl.patch_request_final(ckan)

    print("Update done.")


if __name__ == '__main__':
    ckan = CkanApi(None)
    ckan.initialize_from_cli_args()
    run(ckan)





#!python3
# -*- coding: utf-8 -*-
"""

"""
from typing import List, Dict, Union
import copy
from warnings import warn

from ckanapi_harvesters.auxiliary.proxy_config import ProxyConfig
from ckanapi_harvesters.auxiliary.ckan_api_key import CkanApiKey
from ckanapi_harvesters.auxiliary.ckan_auxiliary import RequestType, assert_or_raise
from ckanapi_harvesters.auxiliary.ckan_map import CkanMap
from ckanapi_harvesters.auxiliary.ckan_errors import MandatoryAttributeError
from ckanapi_harvesters.auxiliary.ckan_vocabulary_deprecated import CkanTagVocabularyInfo, CkanVocabularyMap
from ckanapi_harvesters.policies.data_format_policy import CkanPackageDataFormatPolicy
from ckanapi_harvesters.policies.data_format_policy_tag_groups import TagListPolicy
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_abc import CkanDataCleanerABC

from ckanapi_harvesters.ckan_api.ckan_api import CkanApiParams
from ckanapi_harvesters.ckan_api.deprecated.ckan_api_deprecated import CkanApiDeprecated


class CkanApiVocabulariesDeprecated(CkanApiDeprecated):
    def __init__(self, url:str=None, *, proxies:Union[str,dict,ProxyConfig]=None,
                 apikey:Union[str,CkanApiKey]=None, apikey_file:str=None,
                 owner_org:str=None, params:CkanApiParams=None,
                 map:CkanMap=None, policy: CkanPackageDataFormatPolicy = None, policy_file:str=None,
                 data_cleaner_upload:CkanDataCleanerABC=None,
                 identifier=None):
        """
        CKAN Database API interface to CKAN server with helper functions using pandas DataFrames.

        :param url: url of the CKAN server
        :param proxies: proxies to use for requests
        :param ckan_headers: headers to use for requests, only to the CKAN server
        :param http_headers: headers to use for requests, for all requests, including external requests and to the CKAN server
        :param apikey: way to provide the API key directly (optional)
        :param apikey_file: path to a file containing a valid API key in the first line of text (optional)
        :param policy: data format policy to use with policy_check function
        :param policy_file: path to a JSON file containing the data format policy to use with policy_check function
        :param owner_org: name of the organization to limit package_search (optional)
        """
        msg = DeprecationWarning("Vocabularies are used to define custom fields which accept specific values and require to implement an IDatasetForm extension. This is not covered in this package.")
        warn(msg)
        super().__init__(url=url, proxies=proxies, apikey=apikey, apikey_file=apikey_file,
                         owner_org=owner_org, policy=policy, policy_file=policy_file, identifier=identifier,
                         params=params, map=map, data_cleaner_upload=data_cleaner_upload)
        self.map_vocabulary: CkanVocabularyMap = CkanVocabularyMap()

    def copy(self, new_identifier: str = None, *, dest=None):
        if dest is None:
            dest = CkanApiVocabulariesDeprecated()
        super().copy(new_identifier=new_identifier, dest=dest)
        dest.map_vocabulary = self.map_vocabulary.copy()
        return dest

    def set_default_map_mode(self, datastore_info:bool=None, resource_view_list:bool=None,
                             organization_info:bool=None, license_list:bool=None,
                             load_policy:bool=None, vocabulary_list:bool=None) -> None:
        super().set_default_map_mode(datastore_info=datastore_info, resource_view_list=resource_view_list,
                                     organization_info=organization_info, license_list=license_list,
                                     load_policy=load_policy)
        if vocabulary_list is None:
            vocabulary_list = self.map_vocabulary._mapping_query_vocabulary_list
        self.map_vocabulary._mapping_query_vocabulary_list = vocabulary_list

    def map_resources(self, package_list:Union[str, List[str]]=None, *, params:dict=None,
                      datastore_info:bool=None, resource_view_list:bool=None, organization_info:bool=None, license_list:bool=None,
                      only_missing:bool=True, error_not_found:bool=True,
                      owner_org:str=None, load_policy:bool=None, vocabulary_list:bool=None) -> CkanMap:
        # overload including a call to load the default data format policy
        self.set_default_map_mode(vocabulary_list=vocabulary_list)
        map = super().map_resources(package_list=package_list, params=params, datastore_info=datastore_info,
                              resource_view_list=resource_view_list, organization_info=organization_info,
                              license_list=license_list, only_missing=only_missing, error_not_found=error_not_found,
                              owner_org=owner_org, load_policy=load_policy)
        vocabulary_list = self.map_vocabulary._mapping_query_vocabulary_list
        if vocabulary_list:
            self.vocabulary_list(cancel_if_present=True)
        return map


    ## Vocabulary management (requires sysadmin rights) --------------
    def _api_vocabulary_list(self, *, params:dict=None) -> List[CkanTagVocabularyInfo]:
        """
        API call to vocabulary_list.

        :return: a list of vocabulary info objects
        """
        msg = DeprecationWarning("Vocabulary functions did not work when tested")
        warn(msg)
        response = self._api_action_request(f"vocabulary_list", method=RequestType.Post, json=params)
        if response.success:
            vocabulary_list = [CkanTagVocabularyInfo(vocabulary_dict) for vocabulary_dict in response.result]
            self.map_vocabulary._update_vocabulary_info(vocabulary_list, vocabularies_listed=True)  # update map
            return copy.deepcopy(vocabulary_list)
        else:
            raise response.default_error(self)

    def vocabulary_list(self, cancel_if_present:bool=True) -> List[CkanTagVocabularyInfo]:
        if self.map_vocabulary.vocabularies_listed and cancel_if_present:
            return list(self.map_vocabulary.vocabularies.values())
        else:
            return self._api_vocabulary_list()

    def _api_vocabulary_create(self, vocabulary_name: str, tags_list_dict: List[Dict[str, str]], *, params:dict=None) -> CkanTagVocabularyInfo:
        """
        API call to vocabulary_create.

        :return: a
        """
        msg = DeprecationWarning("Vocabulary functions did not work when tested")
        warn(msg)
        if params is None: params = {}
        params["name"] = vocabulary_name
        params["tags"] = tags_list_dict
        response = self._api_action_request(f"vocabulary_create", method=RequestType.Post, json=params)
        if response.success:
            vocabulary_info = CkanTagVocabularyInfo(response.result)
            self.map_vocabulary._update_vocabulary_info(vocabulary_info)
            return copy.deepcopy(vocabulary_info)
        else:
            raise response.default_error(self)

    def _api_vocabulary_update(self, vocabulary_id: str, tags_list_dict: List[Dict[str, str]], *, params:dict=None) -> CkanTagVocabularyInfo:
        """
        API call to vocabulary_update.

        :return: a
        """
        msg = DeprecationWarning("Vocabulary functions did not work when tested")
        warn(msg)
        if params is None: params = {}
        params["id"] = vocabulary_id
        params["tags"] = tags_list_dict
        response = self._api_action_request(f"vocabulary_update", method=RequestType.Post, json=params)
        if response.success:
            vocabulary_info = CkanTagVocabularyInfo(response.result)
            self.map_vocabulary._update_vocabulary_info(vocabulary_info)
            return copy.deepcopy(vocabulary_info)
        else:
            raise response.default_error(self)

    def vocabulary_update(self, vocabulary_name: str, tags_list_dict: List[Dict[str, str]]):
        vocabulary_id = self.map_vocabulary.get_vocabulary_id(vocabulary_name, error_not_mapped=False)
        if vocabulary_id is None:
            self._api_vocabulary_create(vocabulary_name=vocabulary_name, tags_list_dict=tags_list_dict)
        else:
            self._api_vocabulary_update(vocabulary_id, tags_list_dict=tags_list_dict)

    def _api_vocabulary_delete(self, vocabulary_id: str, *, params:dict=None) -> bool:
        """
        API call to vocabulary_delete.

        :return: True if success
        """
        msg = DeprecationWarning("Vocabulary functions did not work when tested")
        warn(msg)
        if params is None: params = {}
        params["id"] = vocabulary_id
        response = self._api_action_request(f"vocabulary_delete", method=RequestType.Post, json=params)
        if response.success:
            return True
        else:
            raise response.default_error(self)

    def vocabulary_delete(self, vocabulary_id: str) -> bool:
        return self._api_vocabulary_delete(vocabulary_id)

    def vocabularies_clear(self):
        self.vocabulary_list(cancel_if_present=True)
        vocabulary_ids = list(self.map_vocabulary.vocabularies.keys())
        for vocabulary_id in vocabulary_ids:
            self._api_vocabulary_delete(vocabulary_id)

    def initiate_vocabularies_from_policy(self, policy:CkanPackageDataFormatPolicy, *, remove_others:bool=False):
        vocabulary_policy = policy.package_tags
        vocabulary_list: TagListPolicy
        vocabulary_names = {vocabulary_list.group_name for vocabulary_list in vocabulary_policy.value_group_specs}
        if remove_others:
            self.vocabulary_list(cancel_if_present=True)
            current_vocabularies = set(self.map_vocabulary.vocabulary_id_index.keys())
            extra_vocabularies = current_vocabularies - vocabulary_names
            for vocabulary_name in extra_vocabularies:
                vocabulary_id = self.map_vocabulary.vocabulary_id_index[vocabulary_name]
                self._api_vocabulary_delete(vocabulary_id)
        for vocabulary_list in vocabulary_policy.value_group_specs:
            vocabulary_name = vocabulary_list.group_name
            tags_list_dict = vocabulary_list.get_tags_list_dict()
            assert_or_raise(vocabulary_name is not None, MandatoryAttributeError("Tag vocabulary", "vocabulary_name"))
            self.vocabulary_update(vocabulary_name=vocabulary_name, tags_list_dict=tags_list_dict)



#!python3
# -*- coding: utf-8 -*-
"""
Alias to most complete BuilderPackage implementation
"""

from ckanapi_harvesters.builder.builder_package_1_basic import example_package_xls, example_package_resources_dir
from ckanapi_harvesters.builder.builder_package_2_harvesters import BuilderPackageWithHarvesters as BuilderPackage  # alias




#!python3
# -*- coding: utf-8 -*-
"""
Progress callback function interface
"""
from typing import Any, Union, Callable, Dict
from abc import ABC, abstractmethod
from enum import IntEnum
import copy
import time
from warnings import warn


class CkanCallbackLevel(IntEnum):
    Packages = 0
    Resources = 1
    ResourceChunks = 2
    MultiFileResource = 3
    Requests = 4


class CkanProgressBarType(IntEnum):
    NoBar = 0
    TqdmAuto = 1
    TqdmConsole = 2
    TqdmJupyter = 3


class CkanProgressCallbackABC(ABC):
    default_progress_bar_type = CkanProgressBarType.NoBar

    def __init__(self, callback_fun:Union[Callable,"CkanProgressCallbackABC"]=None,
                 *, progress_bar_type:CkanProgressBarType = None):
        if progress_bar_type is None:
            progress_bar_type = self.default_progress_bar_type
        self.progress_callback_fun: Union[Callable[[int, int, Any], None], None] = None
        self.progress_callback_kwargs: dict = {}
        self.verbosity:dict[CkanCallbackLevel,bool] = {level: True for level in CkanCallbackLevel}
        # self.verbosity[CkanCallbackLevel.Request] = False
        self.progress_bar_enables:dict[CkanCallbackLevel,bool] = {level: False for level in CkanCallbackLevel}
        self._progress_bar_type: CkanProgressBarType = CkanProgressBarType.NoBar
        self.progress_bar_type: CkanProgressBarType = progress_bar_type
        # state variables
        self.creation_time = time.time()
        self.start_time:Dict[CkanCallbackLevel,Union[float,None]] = {level: None for level in CkanCallbackLevel}
        self.progress_bars: Dict[CkanCallbackLevel, Any] = {level: None for level in CkanCallbackLevel}
        self.last_progress_position:dict[CkanCallbackLevel,int] = {level: 0 for level in CkanCallbackLevel}
        self.last_progress_file_index:dict[CkanCallbackLevel,int] = {level: 0 for level in CkanCallbackLevel}
        if isinstance(callback_fun, CkanProgressCallbackABC):
            callback_fun.copy(dest=self)
        else:
            self.progress_callback_fun = callback_fun

    @property
    def progress_bar_type(self) -> CkanProgressBarType:
        return self._progress_bar_type
    @progress_bar_type.setter
    def progress_bar_type(self, value: CkanProgressBarType):
        if not value == CkanProgressBarType.NoBar:
            msg = f"Cannot use progress bar type {value.name} for this implementation"
            warn(msg)
        self._progress_bar_type = value

    @abstractmethod
    def copy(self, *, dest=None):
        dest.progress_callback_fun = self.progress_callback_fun
        dest.progress_callback_kwargs = copy.deepcopy(self.progress_callback_kwargs)
        dest.verbosity = self.verbosity.copy()
        dest.progress_bar_enables = self.progress_bar_enables.copy()
        dest._progress_bar_type = self._progress_bar_type
        # do not copy state variables: creation_time and start_time
        return dest

    def __copy__(self):
        return self.copy()

    def start_task(self, total:int, *, file_count:int=None, position:int=0, file_index:int=0, level:CkanCallbackLevel=None,
                 info:Any=None, context:str=None, lines_chunk:int=None, total_lines_read:int=0, **kwargs) -> None:
        if level is not None:
            self.start_time[level] = time.time()
            self.last_progress_position[level] = 0
            self.last_progress_file_index[level] = 0
        self.task_progress(position=position, total=total, file_index=file_index, file_count=file_count, level=level,
            info=info, context=context, lines_chunk=lines_chunk, total_lines_read=total_lines_read,
            canceled_request=False, end_message=False, **kwargs)

    def end_task(self, total:int, *, file_count:int=None, position:int=None, file_index:int=None, level:CkanCallbackLevel=None,
                 info:Any=None, context:str=None, lines_chunk:int=None, total_lines_read:int=None, **kwargs) -> None:
        if position is None:
            position = total
        if file_index is None:
            file_index = file_count
        self.task_progress(position=position, total=total, file_index=file_index, file_count=file_count, level=level,
            info=info, context=context, lines_chunk=lines_chunk, total_lines_read=total_lines_read,
            canceled_request=False, end_message=True, **kwargs)

    @abstractmethod
    def task_progress(self, position:int, total:int, *, info:Any=None, context:str=None,
        file_index:int=0, file_count:int=None, lines_chunk:int=None, total_lines_read:int=None,
        canceled_request: bool=False, end_message: bool=False, level:CkanCallbackLevel=None,
        **kwargs) -> Union[str,None]:
        raise NotImplementedError()



#!python3
# -*- coding: utf-8 -*-
"""
Progress callback function definition
"""
from typing import Any

from ckanapi_harvesters.auxiliary.ckan_progress_callbacks import CkanCallbackLevel


# from ipywidgets import IntProgress
# from IPython.display import display
# f = IntProgress(min=0,max=100)
f = None

def jupyter_progress_callback(position:int, total:int, info:Any=None, *, context:str=None,
                              file_index:int=None, file_count:int=None,
                              lines_chunk:int=None, total_lines_read:int=None,
                              canceled_upload: bool=False, end_message: bool=False,
                              level:CkanCallbackLevel=None, start_time: float=None,
                              last_position:int=None, last_progress_position:int=None,
                              **kwargs) -> None:
    """
    Example of a progress_callback function which can be copied into a Jupyter Notebook using a progress bar:
    ```python
    from ipywidgets import IntProgress
    from IPython.display import display
    f = IntProgress(min=0,max=100)
    ```
    """
    if level == CkanCallbackLevel.ResourceChunks:
        f.value = int(position/total*100)




#!python3
# -*- coding: utf-8 -*-
"""
Progress callback function definition
"""

from ckanapi_harvesters.auxiliary.ckan_progress_callbacks_abc import CkanCallbackLevel, CkanProgressBarType, CkanProgressCallbackABC
from ckanapi_harvesters.auxiliary.ckan_progress_callbacks_simple import CkanProgressCallbackSimple, default_progress_callback
from ckanapi_harvesters.auxiliary.ckan_progress_callbacks_tqdm import CkanProgressCallbackTqdm, tqdm

if tqdm is None:  # ImportError
    CkanProgressCallback = CkanProgressCallbackSimple
else:
    CkanProgressCallback = CkanProgressCallbackTqdm






#!python3
# -*- coding: utf-8 -*-
"""
Progress callback function definition
"""
from typing import Any, Union, Callable
from threading import Semaphore

from ckanapi_harvesters.auxiliary.ckan_progress_callbacks_abc import CkanCallbackLevel, CkanProgressBarType
from ckanapi_harvesters.auxiliary.ckan_progress_callbacks_simple import CkanProgressCallbackSimple
from ckanapi_harvesters.auxiliary.ckan_errors import RequirementError

try:
    from tqdm import tqdm
    from tqdm.autonotebook import tqdm as tqdm_auto
    from tqdm.notebook import tqdm as tqdm_notebook
except ImportError:
    tqdm = None
    tqdm_auto = None
    tqdm_notebook = None


class CkanProgressCallbackTqdm(CkanProgressCallbackSimple):
    default_progress_bar_type = CkanProgressBarType.TqdmAuto
    progress_bar_update_threshold_pct = 0.5

    def __init__(self, callback_fun:Union[Callable,"CkanProgressCallbackSimple"]=None,
                 *, progress_bar_type:CkanProgressBarType=None):
        super().__init__(callback_fun, progress_bar_type=progress_bar_type)
        progress_bar_type = self.progress_bar_type  # update
        self.progress_bar_enables[CkanCallbackLevel.ResourceChunks] = True  # activate this one
        # self.progress_bar_enables[CkanCallbackLevel.Requests] = True  # activate this one
        if False:
            if not progress_bar_type == CkanProgressBarType.NoBar:
                for level in CkanCallbackLevel:
                    if self.progress_bar_enables[level]:
                        self.verbosity[level] = False
        if tqdm is None and not progress_bar_type == CkanProgressBarType.NoBar:
            raise RequirementError('tqdm', "CkanProgressCallbackTqdm")
        # state variables
        self.last_progress_displayed:dict[CkanCallbackLevel,int] = {level: 0 for level in CkanCallbackLevel}
        self.tqdm_semaphore = Semaphore()

    @property
    def progress_bar_type(self) -> CkanProgressBarType:
        return self._progress_bar_type
    @progress_bar_type.setter
    def progress_bar_type(self, value: CkanProgressBarType):
        self._progress_bar_type = value

    def copy(self, *, dest=None):
        if dest is None:
            dest = CkanProgressCallbackTqdm()
        return super().copy(dest=dest)

    def start_task(self, total:int, *, file_count:int=None, position:int=0, file_index:int=0, level:CkanCallbackLevel=None,
                 info:Any=None, context:str=None, lines_chunk:int=None, total_lines_read:int=0, **kwargs) -> None:
        super().start_task(total=total, file_count=file_count, level=level, position=position, file_index=file_index,
                           info=info, context=context, lines_chunk=lines_chunk, total_lines_read=total_lines_read, **kwargs)
        if (level is not None and total is not None
                and not self._progress_bar_type == CkanProgressBarType.NoBar
                and self.progress_bar_enables[level]):
            if self.progress_bars[level] is None:
                self.last_progress_displayed[level] = 0
                desc = level.name
                if context:
                    desc = desc + " (" + context + ")"
                if self._progress_bar_type == CkanProgressBarType.TqdmAuto:
                    self.progress_bars[level] = tqdm_auto(total=total, unit="U", unit_scale=True, desc=desc)  #, position=int(level))
                elif self._progress_bar_type == CkanProgressBarType.TqdmConsole:
                    self.progress_bars[level] = tqdm(total=total, unit="U", unit_scale=True, desc=desc)  #, position=int(level))
                elif self._progress_bar_type == CkanProgressBarType.TqdmJupyter:
                    self.progress_bars[level] = tqdm_notebook(total=total, unit="U", unit_scale=True, desc=desc)  #, position=int(level))
                else:
                    raise NotImplementedError(self._progress_bar_type.name)
                if position > 0:
                    self.progress_bars[level].update(position)
                    self.last_progress_displayed[level] = position

    def end_task(self, total:int, *, file_count:int=None, position:int=None, file_index:int=None, level:CkanCallbackLevel=None,
                 info:Any=None, context:str=None, lines_chunk:int=None, total_lines_read:int=None, **kwargs) -> None:
        super().end_task(total=total, position=position, file_index=file_index, file_count=file_count, level=level,
                         info=info, context=context, lines_chunk=lines_chunk, total_lines_read=total_lines_read, **kwargs)
        if level is not None:
            if self.progress_bars[level] is not None:
                if total is not None:
                    delta = max(0, total - self.progress_bars[level].last_print_n)
                    self.progress_bars[level].update(delta)
                    self.last_progress_displayed[level] += delta
                self.progress_bars[level].close()
                self.progress_bars[level] = None

    def task_progress(self, position:int, total:int, *, info:Any=None, context:str=None,
        file_index:int=0, file_count:int=None, lines_chunk:int=None, total_lines_read:int=None,
        canceled_request: bool=False, end_message: bool=False, level:CkanCallbackLevel=None,
        **kwargs) -> Union[str,None]:
        self.tqdm_semaphore.acquire()
        # if level is not None:
        #     # last_position = self.last_progress_position[level]
        #     last_position = self.progress_bars[level].last_print_n
        msg = super().task_progress(position=position, total=total, file_index=file_index, file_count=file_count, level=level,
                                    context=context, lines_chunk=lines_chunk, total_lines_read=total_lines_read,
                                    canceled_request=canceled_request, end_message=end_message, info=info, **kwargs)
        if level is not None and position is not None:
            if self.progress_bars[level] is not None:
                if total < 0: total = 1  # avoid division by zero
                delta = position - self.progress_bars[level].last_print_n
                if position >= total:  # or end_message:
                    delta = total - self.progress_bars[level].last_print_n
                    update_needed = True  # end of progress
                else:
                    delta_pct = delta / total * 100.
                    update_needed = delta_pct > self.progress_bar_update_threshold_pct
                if update_needed:
                    self.progress_bars[level].update(delta)
                    self.last_progress_displayed[level] += delta
        self.tqdm_semaphore.release()
        return msg

    def __del__(self) -> None:
        for progress_bar in self.progress_bars.values():
            if progress_bar is not None:
                progress_bar.close()



#!python3
# -*- coding: utf-8 -*-
"""
Code to initiate a package builder from a Dataset harvester
"""
from typing import List

from ckanapi_harvesters.builder.builder_package_1_basic import BuilderPackageBasic
from ckanapi_harvesters.builder.builder_resource_datastore_multi_harvester import BuilderDataStoreHarvester
from ckanapi_harvesters.harvesters.harvester_abc import DatasetHarvesterABC


class BuilderPackageWithHarvesters(BuilderPackageBasic):
    @staticmethod
    def init_from_harvester(dataset_harvester: DatasetHarvesterABC) -> "BuilderPackageWithHarvesters":
        builder = BuilderPackageWithHarvesters()
        params = dataset_harvester.params
        builder.package_name = f"harvest_{params.harvest_method}_{params.database}_{params.dataset}".lower()
        builder.package_attributes.title = f"Automated harvesting of a {params.harvest_method} dataset: database {params.database} / dataset {params.dataset} [PoC]"
        builder.package_attributes.description = f"Harvested from {params.host} / database {params.database} / dataset {params.dataset}. Testing the upload of a full dataset from {params.harvest_method}."
        builder.package_attributes.private = True
        # builder.package_attributes.state = CkanState.Draft
        tables: List[str] = dataset_harvester.list_tables(return_metadata=False)
        for table_name in tables:
            table_harvester = dataset_harvester.get_table_harvester(table_name)
            resource_builder = BuilderDataStoreHarvester(parent=builder)
            resource_builder.name = table_name
            resource_builder.harvester = table_harvester
            if resource_builder.resource_attributes_user.description is None:
                resource_builder.resource_attributes_user.description = f"dataset {params.dataset} / table {table_name}"
            # metadata is imported after a clean of metadata
            builder.resource_builders[table_name] = resource_builder
        return builder

    def copy(self, dest=None) -> "BuilderPackageWithHarvesters":
        if dest is None:
            dest = BuilderPackageWithHarvesters()
        super().copy(dest=dest)
        return dest



#!python3
# -*- coding: utf-8 -*-
"""
The basic file format for DataStore: XLS
"""
from typing import Union, Dict
import io
import argparse

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_model import CkanField
from ckanapi_harvesters.auxiliary.list_records import ListRecords
from ckanapi_harvesters.harvesters.file_formats.file_format_abc import FileFormatABC


class ExcelFileFormat(FileFormatABC):
    """
    Excel file IO using pandas.read_excel. Supports xls, xlsx, xlsm, xlsb, odf, ods and odt formats.

    Recommended: use with CLI argument `--read-kwargs sheet_name=your_sheet` (default is 0 i.e. the first sheet)
    """
    def __init__(self, options_string: str, *, read_kwargs: dict=None, write_kwargs: dict=None) -> None:
        super().__init__(options_string=options_string, read_kwargs=read_kwargs, write_kwargs=write_kwargs)
        self.allow_chunks = False  # override: reading by chunks is not possible

    @staticmethod
    def _setup_cli_parser(parser: argparse.ArgumentParser = None) -> argparse.ArgumentParser:
        parser = FileFormatABC._setup_cli_parser(parser)
        parser.add_argument("--sheet-name", type=str,
                            help="The name of the sheet to use (defaults to first sheet)")
        return parser

    def _apply_arguments(self, args: argparse.Namespace, extra_args: list):
        super()._apply_arguments(args, extra_args)
        if args.sheet_name is not None:
            self.read_kwargs["sheet_name"] = args.sheet_name
            self.write_kwargs["sheet_name"] = args.sheet_name

    # read -------------------
    def read_file(self, file_path: str, fields: Union[Dict[str, CkanField],None], allow_chunks:bool=True) -> Union[pd.DataFrame, ListRecords]:
        read_kwargs = self._get_read_kwargs(allow_chunks=False)
        return pd.read_excel(file_path, **read_kwargs)

    def read_buffer_full(self, buffer: io.StringIO, fields: Union[Dict[str, CkanField],None]) -> Union[pd.DataFrame, ListRecords]:
        read_kwargs = self._get_read_kwargs(allow_chunks=False)
        return pd.read_excel(buffer, **read_kwargs)

    # write ------------------
    def write_file(self, df: pd.DataFrame, file_path: str, fields: Union[Dict[str, CkanField],None]) -> None:
        write_kwargs = self._get_write_kwargs()
        df.to_excel(file_path, index=False, **write_kwargs)

    def write_in_memory(self, df: pd.DataFrame, fields: Union[Dict[str, CkanField],None]) -> bytes:
        with io.BytesIO() as buffer:
            write_kwargs = self._get_write_kwargs()
            df.to_excel(buffer, index=False, **write_kwargs)
            return buffer.getvalue()

    def append_allowed(self) -> bool:
        return False

    def append_file(self, df: Union[pd.DataFrame, ListRecords], file_path: str,
                    fields: Union[Dict[str, CkanField], None]) -> None:
        raise NotImplementedError()

    def append_in_memory(self, stream: bytes, df: Union[pd.DataFrame, ListRecords], fields: Union[Dict[str, CkanField],None]) -> bytes:
        raise NotImplementedError()

    # misc ------------------
    def copy(self, dest=None):
        if dest is None:
            dest = ExcelFileFormat(self.options_string)
        super().copy(dest=dest)
        return dest






#!python3
# -*- coding: utf-8 -*-
"""
Progress callback function definition
"""
from typing import Any, Union, Callable
from threading import Semaphore

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_auxiliary import FileChunkDataFrame
from ckanapi_harvesters.auxiliary.ckan_progress_callbacks_abc import (CkanCallbackLevel, CkanProgressBarType,
                                                                      CkanProgressCallbackABC)


def default_progress_callback(position:int, total:int, info:Any=None, *, context:str=None,
                              file_index:int=None, file_count:int=None,
                              lines_chunk:int=None, total_lines_read:int=None,
                              canceled_upload: bool=False, end_message: bool=False,
                              level:CkanCallbackLevel=None, start_time: float=None,
                              last_position:int=None, last_progress_position:int=None,
                              **kwargs) -> Union[str,None]:
    if total is not None and total > 0:
        progress_pct_str = "{0:.2f}".format(position/total*100.0)
    else:
        progress_pct_str = None
    msg = None
    if level == CkanCallbackLevel.Packages:
        msg = f"Package {position}/{total}"
    elif level == CkanCallbackLevel.Resources:
        if end_message:
            msg = f"Done all {total} resources"
        else:
            msg = f"Resource {position}/{total}"
    elif level == CkanCallbackLevel.MultiFileResource:
        msg = f"Multi-file resource / file {position}/{total}"
    elif level == CkanCallbackLevel.ResourceChunks:
        if context is None:
            context = ""
        if position == total and end_message:
            # info is None
            msg = f"{context} Finished {file_index}/{file_count} (100%) - {total_lines_read} lines read"
        elif canceled_upload:
            msg = f"{context} Canceled {file_index}/{file_count} ({progress_pct_str}%) - {total_lines_read} lines read"
        elif info is None:
            msg = f"{context} Request {file_index}/{file_count} ({progress_pct_str}%) - {total_lines_read} lines read"
        else:
            if isinstance(info, str):
                info_str = info
            elif isinstance(info, pd.DataFrame):
                if "source" in info.attrs.keys():
                    info_str = str(info.attrs["source"])
                else:
                    info_str = "<DataFrame>"
            elif isinstance(info, list):
                info_str = "<records>"
            elif isinstance(info, FileChunkDataFrame):
                if isinstance(info.df, pd.DataFrame):
                    if "source" in info.df.attrs.keys():
                        info_str = str(info.df.attrs["source"])
                    else:
                        info_str = "<DataFrame>"
                elif isinstance(info.df, list):
                    info_str = "<records>"
                else:
                    info_str = str(info.df)
            else:
                info_str = str(info)
            msg = f"{context} Request {file_index}/{file_count} ({progress_pct_str}%) - {total_lines_read} lines read: " + info_str
    elif level == CkanCallbackLevel.Requests:
        msg = f"Multi-line request {position}/{total} ({progress_pct_str}%)"
        if context is not None:
            msg = msg + f" ({context})"
    return msg


class CkanProgressCallbackSimple(CkanProgressCallbackABC):
    def __init__(self, callback_fun:Union[Callable,"CkanProgressCallbackSimple"]=None,
                 *, progress_bar_type:CkanProgressBarType=None):
        if progress_bar_type is None:
            progress_bar_type = self.default_progress_bar_type
        if callback_fun is None:
            callback_fun = default_progress_callback
        super().__init__(callback_fun, progress_bar_type=progress_bar_type)
        self.simple_semaphore = Semaphore()

    def copy(self, *, dest=None):
        if dest is None:
            dest = CkanProgressCallbackSimple()
        return super().copy(dest=dest)

    def __copy__(self):
        return self.copy()

    def task_progress(self, position:int, total:int, *, info:Any=None, context:str=None,
        file_index:int=0, file_count:int=None, lines_chunk:int=None, total_lines_read:int=None,
        canceled_request: bool=False, end_message: bool=False, level:CkanCallbackLevel=None,
        **kwargs) -> Union[str,None]:
        """
        Progress callback function. Use to implement a progress indication for the user.

        :param position: the position within the resource (usually, in bytes or line count)
        :param total: the total size of the resource
        :param info: an object from which more information can be extracted, typically, the DataFrame itself, with an indication of the data origin.
        :param context: the context of the call (ckan instance, upload/download, single/multi-threaded)
        :param file_index: the index of the file in the list
        :param file_count: the number of files in the list
        :param lines_chunk: the number of lines in the chunk currently being processed
        :param total_lines_read: the total number of lines read, including the current chunk
        :param canceled_request: this callback is also called when a line is ignored
        :param end_message: boolean indicating of the work in progress
        :param level: the level of the progress callback
        """
        self.simple_semaphore.acquire()
        if level is not None:
            last_position = self.last_progress_position[level]
            last_file_index = self.last_progress_file_index[level]
            self.last_progress_position[level] = position
            self.last_progress_file_index[level] = file_index
            if not self.verbosity[level]:
                self.simple_semaphore.release()
                return None
        if self.progress_callback_fun is not None:
            if end_message:
                position = total
                file_index = file_count
            # division by zero error prevention:
            if total == 0: total = max(1, position)
            if file_count is not None and file_count == 0: file_count = max(file_index, 1)
            # timing of the current task
            if level is not None:
                task_start_time = self.start_time[level]
                # if task_start_time is None or position == 0 or end_message:
                #     task_start_time = time.time()
                #     self.start_time[level] = task_start_time
            else:
                task_start_time = None
            # call user-defined function
            print_msg = self.progress_callback_fun(position, total, info=info, context=context,
                                              file_index=file_index, file_count=file_count,
                                              lines_chunk=lines_chunk, total_lines_read=total_lines_read,
                                              canceled_upload=canceled_request, end_message=end_message,
                                              level=level, start_time=task_start_time,
                                              last_position=last_position, last_file_index=last_file_index,
                                              **self.progress_callback_kwargs, **kwargs)
            self.simple_semaphore.release()
            if print_msg is not None:
                print(print_msg)
            return print_msg
        else:
            self.simple_semaphore.release()



#!python3
# -*- coding: utf-8 -*-
"""
Example code to upload the builder example to a CKAN server
"""
import os

from ckanapi_harvesters.builder.builder_package import BuilderPackage
from ckanapi_harvesters.ckan_api import CkanApi

from ckanapi_harvesters.builder.example import example_package_xls


def run(ckan:CkanApi = None):
    BuilderPackage.unlock_external_code_execution()

    mdl = BuilderPackage.from_excel(example_package_xls)
    ckan = mdl.init_ckan(ckan)
    ckan.input_missing_info(input_args_if_necessary=True, input_owner_org=True)
    ckan.set_limits(10000)  # reduce if server hangs up
    ckan.set_submit_timeout(5)
    ckan.set_verbosity(True)
    ckan.test_ckan_login(raise_error=True, verbose=True)
    # ckan.set_default_map_mode(datastore_info=True)  # uncomment to query DataStore information

    # Test re-encoding the Excel file from the loaded model
    example_package_xls_reencoded = os.path.abspath("builder_package_example-reencoded.xlsx")
    mdl.to_excel(example_package_xls_reencoded)

    # Patch package: apply metadata and upload small resources
    reupload = True  # True: reuploads all documents and resets large datasets to the first document
    mdl.patch_request_full(ckan, reupload=reupload)

    # Upload large datasets
    print("upload_large_datasets...")
    threads = 1  # > 1: multi-threading mode - reduce if HTTP 502 errors
    mdl.upload_large_datasets(ckan, threads=threads)

    print("Update done.")


if __name__ == '__main__':
    ckan = CkanApi(None)
    ckan.initialize_from_cli_args()
    run(ckan)




#!python3
# -*- coding: utf-8 -*-
"""
Example code to download the builder example from a CKAN server
"""
import os

from ckanapi_harvesters.builder.builder_package import BuilderPackage
from ckanapi_harvesters.ckan_api import CkanApi

from ckanapi_harvesters.builder.example import example_package_xls
example_package_download_dir = os.path.abspath("package_download")


def run(ckan:CkanApi = None):
    BuilderPackage.unlock_external_code_execution()

    mdl = BuilderPackage.from_excel(example_package_xls)
    ckan = mdl.init_ckan(ckan)
    ckan.input_missing_info(input_args_if_necessary=True, input_owner_org=True)
    ckan.set_verbosity(True)

    # download into example_package_download_dir
    threads = 3  # > 1: number of threads to download large datasets
    mdl.download_request_full(ckan, example_package_download_dir, full_download=True, threads=threads,
                              skip_existing=False, rm_dir=True)

    print("Package downloaded in")
    print(example_package_download_dir)


if __name__ == '__main__':
    ckan = CkanApi(None)
    ckan.initialize_from_cli_args()
    run(ckan)






#!python3
# -*- coding: utf-8 -*-
"""
Data model to represent a CKAN database architecture
"""
import datetime
from abc import ABC, abstractmethod
from enum import IntEnum, IntFlag
from typing import List, Dict, Union
from warnings import warn
import copy
from collections import OrderedDict

from ckanapi_harvesters.auxiliary.ckan_auxiliary import assert_or_raise, _bool_from_string, bytes_to_megabytes
from ckanapi_harvesters.auxiliary.ckan_auxiliary import CkanFieldInternalAttrs
from ckanapi_harvesters.auxiliary.ckan_errors import IntegrityError, MissingIdError
from ckanapi_harvesters.auxiliary.ckan_auxiliary import dict_recursive_update

## Enumerations ------------------
class UpsertChoice(IntEnum):
    Insert = 1
    Update = 2
    Upsert = 3

    def __str__(self):
        return self.name.lower()


class CkanFieldTypeABC(ABC):
    @staticmethod
    @abstractmethod
    def from_str(s):
        raise NotImplementedError()

class CkanFieldType(str, CkanFieldTypeABC):
    """
    Role previously managed by CkanFieldTypeEnum, but accepts any string
    """
    @staticmethod
    def from_str(s):
        return CkanFieldType(s)


class CkanFieldTypeEnum(IntEnum):  #, CkanFieldTypeABC):
    """
    Enumeration of types encountered during development + documentation
    """
    # CKAN web interface
    Text = 1
    Numeric = 2
    TimeStamp = 3
    # [CKAN documentation](https://docs.ckan.org/en/2.9/maintaining/datastore.html#field-types)
    json = 30
    date = 11  # This type is used to represent a calendar date (year, month, day). The oldest date that can be represented is 4713 BC and the latest date is 5874897 AD. The resolution is 1 day.
    time = 12  # This type is used to represent a time of day without time zone. The low value is 00:00:00 and the high value is 24:00:00 with a resolution of 1 microsecond.
    int = 13
    float = 14
    bool = 15
    # [Postgre documentation](https://www.postgresql.org/docs/9.1/datatype.html)
    timetz = 20  # time of day, including time zone
    timestamptz = 21  # date and time, including time zone
    # structured data
    jsonb = 31
    xml = 32
    # numeric types
    int2 = 101
    integer = int  # alias
    int4 = int  # alias
    int8 = 102
    bigint = int8  # alias
    int16 = 103
    int32 = 104
    int64 = 105
    float4 = 106  # real
    float8 = 107  # double
    money = 108
    # decimal [ (p, s) ] : exact numeric of selectable precision
    # other types
    bit = 200  # fixed-length bit string
    char = 201  # This type is used to represent fixed-length, space padded strings of the specified width. Storing character strings longer than the specified length will result in an error unless the excess characters are spaces, in which case the string will be truncated to the maximum length. If the string to be stored is shorter than the declared length, the value will be space padded.
    varbit = 202  # variable-length
    varchar = 203  # variable-length
    bytea = 204  # This type is used to represent binary strings (a “byte array”). A binary string is a sequence of octets (or bytes). Unlike character strings, binary strings allow storing octets of value zero and other non-printable octets (outside the range 32 to 126).
    # identifiers
    serial4 = 150  # This type is used to represent an auto incrementing four-byte integer. The range is 1 to 2147483647. This is similar to specifying an integer column that has default values to be assigned from a sequence generator. It also has a NOT NULL constraint applied to it.
    serial = serial4  # alias
    serial8 = 151
    bigserial = serial8  # alias
    uuid = 152  # This type is used to represent Universally Unique Identifiers (UUIDs). These identifiers are 128-bit values generated by an algorithm. A UUID is a sequence of lower-case hexadecimal digits in several groups separated by hyphens. Specifically, it is a group of 8 digits, followed by three groups of 4 digits, followed by a group of 12 digits. An example of a UUID in this standard form is a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a11.
    oid = 153  #  numeric object identifier. It is currently implemented as an unsigned four-byte integer. Its use as a primary key in a user-created table is discouraged. OIDs are best used only for references to system tables.
    macaddr = 154
    inet = 155
    cidr = 156
    # geometries, on a plane
    point = 220
    path = 221
    polygon = 222
    box = 223
    circle = 224
    lseg = 225  # line segment
    line = 226
    # PostGRE extensions:
    geometry = 228  # PostGIS
    bson = 230

    def __str__(self):
        return self.name.lower()

    @staticmethod
    def from_str(s):
        s = s.lower().strip()
        if s == "text":
            return CkanFieldTypeEnum.Text
        elif s == "numeric":
            return CkanFieldTypeEnum.Numeric
        elif s == "timestamp":
            return CkanFieldTypeEnum.TimeStamp
        # elif s == "" or ((not isinstance(s, str)) and np.isnan(s)):
        #     return CkanFieldTypeEnum.Default
        elif hasattr(CkanFieldTypeEnum, s):  # other attributes are lower case
            return getattr(CkanFieldTypeEnum, s)
        else:
            raise ValueError(s)


class CkanState(IntEnum):
    Draft = 0
    Active = 1
    Deleted = 2

    def __str__(self):
        return self.name.lower()

    @staticmethod
    def from_str(s):
        s = s.lower().strip()
        if s == "active":
            return CkanState.Active
        elif s == "draft":
            return CkanState.Draft
        elif s == "deleted":
            return CkanState.Deleted
        else:
            raise ValueError(s)


class CkanVisibility(IntEnum):
    Private = 0
    Public = 1

    def __str__(self):
        return self.name.lower()

    @staticmethod
    def from_str(s):
        s = s.lower().strip()
        if s == "private":
            return CkanVisibility.Private
        elif s == "public":
            return CkanVisibility.Public
        else:
            raise ValueError(s)

    @staticmethod
    def from_bool_is_private(value):
        if value:
            return CkanVisibility.Private
        else:
            return CkanVisibility.Public

    def to_bool_is_private(self):
        return CkanVisibility.Private == self.value


class CkanLicenseDomain(IntFlag):
    NoDomain = 0
    Software = 1
    Data = 2
    Content = 4

    @staticmethod
    def from_bool(*, domain_software:bool=False, domain_data:bool=False, domain_content:bool=False) -> "CkanLicenseDomain":
        flag = CkanLicenseDomain.NoDomain
        if domain_software:
            flag = flag | CkanLicenseDomain.Software
        if domain_data:
            flag = flag | CkanLicenseDomain.Data
        if domain_content:
            flag = flag | CkanLicenseDomain.Content
        return flag

    def to_dict(self) -> dict:
        return OrderedDict([
                ("domain_software", self.value & CkanLicenseDomain.Software > 0),
                ("domain_data", self.value & CkanLicenseDomain.Data > 0),
                ("domain_content", self.value & CkanLicenseDomain.Content > 0),
            ])

    @staticmethod
    def from_dict(d: dict) -> "CkanLicenseDomain":
        return CkanLicenseDomain.from_bool(domain_software=_bool_from_string(d["domain_software"]),
                                           domain_data=_bool_from_string(d["domain_data"]),
                                           domain_content=_bool_from_string(d["domain_content"]))

class CkanCapacity(IntEnum):
    Excluded = 0
    Member = 1
    Editor = 2  # only for collaborators of a package
    Admin = 3   # only for members of a group
    SysAdmin = 4
    Public = 5  # to notify access is publicly available

    def __str__(self):
        return self.name.lower()

    @staticmethod
    def from_str(s):
        s = s.lower().strip()
        if s == "excluded":
            return CkanCapacity.Excluded
        elif s == "member":
            return CkanCapacity.Member
        elif s == "editor":
            return CkanCapacity.Editor
        elif s == "admin":
            return CkanCapacity.Admin
        elif s == "sysadmin":
            return CkanCapacity.SysAdmin
        elif s == "public":
            return CkanCapacity.Public
        else:
            raise ValueError(s)

class CkanConfigurableObjectABC(ABC):
    mandatory_attributes: set = None
    configurable_attributes: set = None
    extra_attributes: set = set()

    @staticmethod
    @abstractmethod
    def get_resource_type() -> str:
        raise NotImplementedError()


## Field class ------------------
class CkanField(CkanConfigurableObjectABC):
    """
    Object representation of a CKAN Field configuration
    """
    mandatory_attributes = {"name"}
    configurable_attributes = {"name", "notes", "label"}

    # TODO: implement schema part of dict? e.g. {'index_name': None, 'is_index': False, 'native_type': 'numeric', 'notnull': False, 'uniquekey': False}
    def __init__(self, name:str, data_type:str, *, native_type:str=None, notes:str=None,
                 type_override:bool=False, label:str=None):
        if native_type is None:
            native_type = data_type
        self.name:str = name
        self.data_type:Union[CkanFieldType,None] = CkanFieldType.from_str(data_type) if data_type is not None else None
        self.type_override:Union[bool,None] = type_override
        self.label:Union[str,None] = label
        self.notes:Union[str,None] = notes
        self.is_index:Union[bool,None] = None
        self.uniquekey:Union[bool,None] = None
        self.notnull:Union[bool,None] = None
        self.internal_attrs: CkanFieldInternalAttrs = CkanFieldInternalAttrs()
        self.details:dict = {}
        self.details = self.to_ckan_dict()
        self.internal_attrs.init_from_native_type(self.data_type)

    def __str__(self):
        return f"Field '{self.name}' <{self.data_type}>"

    def copy(self) -> "CkanField":
        return copy.deepcopy(self)

    def merge(self, new_values, dest:"CkanField"=None):
        if dest is None:
            dest = self.copy()
        if new_values.name is not None:
            dest.name = new_values.name
        if new_values.data_type is not None:
            dest.data_type = new_values.data_type
        if new_values.type_override is not None:
            dest.type_override = new_values.type_override
        if new_values.label is not None:
            dest.label = new_values.label
        if new_values.notes is not None:
            dest.notes = new_values.notes
        if new_values.is_index is not None:
            dest.is_index = new_values.is_index
        if new_values.uniquekey is not None:
            dest.uniquekey = new_values.uniquekey
        if new_values.notnull is not None:
            dest.notnull = new_values.notnull
        dest.internal_attrs = self.internal_attrs.merge(new_values.internal_attrs)
        dest.details = dict_recursive_update(self.details, new_values.details)
        return dest

    def update_missing(self, other: "CkanField"):
        self.merge(new_values=other, dest=self)

    def __eq__(self, other) -> bool:
        equality = True
        equality &= self.name == other.name
        equality &= self.data_type == other.data_type
        equality &= self.type_override == other.type_override
        equality &= self.label == other.label
        equality &= self.notes == other.notes
        equality &= self.is_index == other.is_index
        equality &= self.uniquekey == other.uniquekey
        equality &= self.notnull == other.notnull
        equality &= self.internal_attrs == other.internal_attrs
        return equality

    @staticmethod
    def get_resource_type() -> str:
        return "Field"

    def to_ckan_dict(self, include_details:bool=True) -> dict:
        d = dict()
        if self.details is not None and include_details:
            d.update(self.details)
        d["id"] = self.name
        if self.data_type is not None:
            d["type"] = str(self.data_type)
        field_info = self.details["info"] if include_details and self.details is not None and "info" in self.details.keys() else {}
        if self.type_override:
            field_info["type_override"] = str(self.data_type)
        if self.label is not None:
            field_info["label"] = self.label
        if self.notes is not None:
            field_info["notes"] = self.notes
        if len(field_info) > 0:
            d["info"] = field_info
        schema_info = self.details["schema"] if include_details and self.details is not None and "schema" in self.details.keys() else {}
        if self.is_index is not None:
            schema_info["is_index"] = self.is_index
        if self.uniquekey is not None:
            schema_info["uniquekey"] = self.uniquekey
        if self.notnull is not None:
            schema_info["notnull"] = self.notnull
        if self.data_type is not None:
            schema_info["native_type"] = str(self.data_type)
        if len(schema_info) > 0:
            d["schema"] = schema_info
        return d

    @staticmethod
    def from_ckan_dict(d:dict) -> "CkanField":
        obj = CkanField(d["id"], d["type"])
        obj.details = d
        if "info" in d.keys():
            field_info = d["info"]
            if "type_override" in field_info.keys():
                if isinstance(field_info["type_override"], str):
                    if len(field_info["type_override"]) > 0:
                        # the API usually returns a string representing the type override, equal to d["type"]
                        if not d["type"] == field_info["type_override"]:
                            obj.data_type = CkanFieldType.from_str(field_info["type_override"])
                            msg = f"Inconsistency between data type and type override for field {obj.name} (data: {d['type']} vs. override: {field_info['type_override']})"
                            warn(msg)
                        obj.type_override = True
                elif isinstance(field_info["type_override"], bool):
                    obj.type_override = field_info["type_override"]
                else:
                    obj.type_override = field_info["type_override"] > 0
            if "label" in field_info.keys():
                obj.label = field_info["label"]
            if "notes" in field_info.keys():
                obj.notes = field_info["notes"]
        if "schema" in d.keys():
            schema_info = d["schema"]
            if "is_index" in schema_info.keys():
                obj.is_index = schema_info["is_index"]
            if "uniquekey" in schema_info.keys():
                obj.uniquekey = schema_info["uniquekey"]
            if "notnull" in schema_info.keys():
                obj.notnull = schema_info["notnull"]
            if "native_type" in schema_info.keys():
                obj.data_type = CkanFieldType.from_str(schema_info["native_type"])
        obj.internal_attrs.init_from_native_type(obj.data_type)
        return obj

    def to_dict(self, include_details:bool=True) -> dict:
        return self.to_ckan_dict(include_details=include_details)

    @staticmethod
    def from_dict(d:dict) -> "CkanField":
        return CkanField.from_ckan_dict(d)


class CkanAliasInfo:
    def __init__(self, d:dict=None):
        self.id: Union[str,None] = None
        self.name: str = ""
        self.alias_of: Union[str,None] = None
        self.details:dict = d
        if d is not None:
            self.id: Union[str, None] = d["id"] if "id" in d.keys() else None
            self.name: str = d["name"]
            self.alias_of: Union[str, None] = d["alias_of"]

    def __str__(self):
        return f"Alias {self.name} of id {self.alias_of}"

    def copy(self) -> "CkanAliasInfo":
        return copy.deepcopy(self)

    @staticmethod
    def from_dict(d:dict) -> "CkanAliasInfo":
        return CkanAliasInfo(d)

    def to_dict(self, include_details:bool=True) -> dict:
        d = dict()
        if self.details is not None and include_details:
            d.update(self.details)
        d.update({"name": self.name, "alias_of": self.alias_of})
        if id is not None:
            d["id"] = id
        return d

## Users and groups ------------------
class CkanUserInfo:
    def __init__(self, d: dict = None):
        self.id: Union[str,None] = None
        self.name: Union[str,None] = None
        self.display_name: Union[str,None] = None
        self.fullname: Union[str,None] = None
        self.about: Union[str,None] = None
        self.sysadmin: bool = False
        self.state: Union[CkanState,None] = None
        self.email_hash: Union[str,None] = None
        self.created: Union[datetime.datetime, None] = None
        self.last_active: Union[datetime.datetime, None] = None
        if d is not None:
            self.id = d["id"]
            self.name = d["name"]
            self.display_name = d["display_name"]
            self.fullname = d["fullname"]
            self.about = d["about"]
            self.sysadmin = d["sysadmin"]
            self.state = CkanState.from_str(d["state"])
            self.email_hash = d["email_hash"] if "email_hash" in d.keys() else None  # MD5 hash
            self.created = datetime.datetime.fromisoformat(d["created"]) if "created" in d.keys() else None
            self.last_active = datetime.datetime.fromisoformat(d["last_active"]) if "last_active" in d.keys() else None
        self.organizations: Union[None,List[str]] = None  # used by consolidate (detailed_report)
        self.details: Union[dict,None] = d

    def __str__(self):
        return f"User '{self.name}' ({self.id})"

    @staticmethod
    def get_resource_type() -> str:
        return "User"

    def to_dict(self, include_details: bool = True) -> dict:
        d = dict()
        if self.details is not None and include_details:
            d.update(self.details)
        d.update({"id": self.id, "name": self.name, "display_name": self.display_name, "fullname": self.fullname,
                  "about": self.about, "sysadmin": self.sysadmin, "state": str(self.state),
                  "created": self.created.isoformat() if self.created is not None else None,
                  "last_active": self.last_active.isoformat() if self.last_active is not None else None})
        return d

    @staticmethod
    def from_dict(d: dict) -> "CkanUserInfo":
        return CkanUserInfo(d)

    def copy(self) -> "CkanUserInfo":
        return copy.deepcopy(self)


class CkanGroupInfo:
    def __init__(self, d:dict):
        self.id:str = d["id"]
        self.name:str = d["name"]
        self.title:str = d["title"]
        self.description:str = d["description"]
        self.package_count:Union[None,int] = d.get("package_count")
        self.details:dict = d
        # to be initialized with specific requests:
        self.user_members:Union[dict[str,CkanCapacity],None] = None
        self.package_members:Union[dict[str,CkanCapacity],None] = None

    def __str__(self):
        return f"Group '{self.title}' ({self.id})"

    @staticmethod
    def get_resource_type() -> str:
        return "Group"

    def to_dict(self, include_details:bool=True) -> dict:
        d = dict()
        if self.details is not None and include_details:
            d.update(self.details)
        d.update({"id": self.id, "name": self.name, "title": self.title, "description": self.description})
        return d

    @staticmethod
    def from_dict(d:dict) -> "CkanGroupInfo":
        return CkanGroupInfo(d)

    def copy(self) -> "CkanGroupInfo":
        return copy.deepcopy(self)


## Package, resources and views map class ------------------
class CkanDataStoreInfo:
    def __init__(self, d:dict=None):
        self.resource_id:Union[str,None] = None
        self.row_count: Union[int,None] = None
        self.fields_id_list:Union[List[str],None] = None
        self.fields_dict:Union[OrderedDict[str,CkanField],None] = None
        self.index_fields:Union[List[str],None] = None
        self.aliases:Union[List[str],None] = None
        self.table_size_mb:Union[float,None] = None
        self.index_size_mb:Union[float,None] = None
        self.details:dict = d
        if d is not None:
            self.resource_id:str = d["meta"]["id"]
            if "aliases" in d["meta"].keys():
                self.aliases = d["meta"]["aliases"]
            if "count" in d["meta"].keys():
                self.row_count:int = d["meta"]["count"]
            self.table_size_mb = bytes_to_megabytes(d["meta"].get("size", None))
            self.index_size_mb = bytes_to_megabytes(d["meta"].get("idx_size", None))
            # what does the field meta.db_size represent?
            if "fields" in d.keys():
                self.fields_id_list:List[str] = [e["id"] for e in d["fields"]]
                self.fields_dict = OrderedDict()
                for e in d["fields"]:
                    self.fields_dict[e["id"]] = CkanField.from_ckan_dict(d=e)
                self.index_fields:List[str] = [e.name for e in self.fields_dict.values() if e.is_index]

    def __str__(self):
        return f"DataStore of resource id {self.resource_id}"

    def get_basic_field_list_dict(self):
        return [{"id": id} for id in self.fields_id_list]

    def get_recomp_field_list_dict(self):
        return [self.fields_dict[id].to_ckan_dict() for id in self.fields_id_list]

    def get_original_field_list_dict(self):
        return [self.fields_dict[id].details for id in self.fields_id_list]

    def copy(self) -> "CkanDataStoreInfo":
        return copy.deepcopy(self)

    def to_dict(self, include_details:bool=True) -> dict:
        d = dict()
        if self.details is not None and include_details:
            d.update(self.details)
        if self.fields_dict is not None:
            d["fields"] = [field.to_dict(include_details=include_details) for field in self.fields_dict.values()]
        if "meta" not in d.keys():
            d["meta"] = {}
        d["meta"]["id"] = self.resource_id
        if self.row_count is not None:
            d["meta"]["count"] = self.row_count
        return d

    @staticmethod
    def from_dict(d:dict) -> "CkanDataStoreInfo":
        return CkanDataStoreInfo(d)


class CkanViewInfo:
    def __init__(self, d:dict):
        self.id:str = d["id"]
        self.title:str = d["title"]
        self.view_type:str = d["view_type"]
        self.resource_id:str = d["resource_id"]
        self.package_id:str = d["package_id"]
        self.details:dict = d

    def __str__(self):
        return f"View '{self.title}' ({self.id} of resource id {self.resource_id})"

    def to_dict(self, include_details:bool=True) -> dict:
        d = dict()
        if self.details is not None and include_details:
            d.update(self.details)
        d.update({"id": self.id, "title": self.title, "view_type": self.view_type, "resource_id": self.resource_id, "package_id": self.package_id})
        return d

    @staticmethod
    def from_dict(d:dict) -> "CkanViewInfo":
        return CkanViewInfo(d)

    def copy(self) -> "CkanViewInfo":
        return copy.deepcopy(self)


class CkanLicenseInfo:
    def __init__(self, d:dict):
        self.id:str = d["id"]
        self.title:str = d["title"]
        self.state:CkanState = CkanState.from_str(d["status"])
        self.family:str = d["family"]
        self.domain:CkanLicenseDomain = CkanLicenseDomain.from_bool(domain_software=_bool_from_string(d["domain_software"]),
                                                        domain_data=_bool_from_string(d["domain_data"]), domain_content=_bool_from_string(d["domain_content"]))
        self.is_generic:Union[bool,None] = _bool_from_string(d["is_generic"]) if "is_generic" in d.keys() else None
        self.url:str = d["url"]
        self.details:dict = d

    def __str__(self):
        return f"License '{self.title}' ({self.id}) [{str(self.domain)}]"

    def to_dict(self, include_details:bool=True) -> dict:
        d = dict()
        if self.details is not None and include_details:
            d.update(self.details)
        d.update({"id": self.id, "title": self.title, "state": str(self.state), "family": self.family})
        d.update(self.domain.to_dict())
        return d

    @staticmethod
    def from_dict(d:dict) -> "CkanLicenseInfo":
        return CkanLicenseInfo(d)


class CkanTagInfo:
    def __init__(self, d:dict):
        self.id:str = d["id"]
        self.name:str = d["name"]
        self.display_name:str = d["display_name"]
        self.state: Union[CkanState,None] = None
        if "state" in d.keys():
            self.state = CkanState.from_str(d["state"])
        self.vocabulary_id:Union[str,None] = d["vocabulary_id"]
        self.details:dict = d

    def __str__(self):
        if self.vocabulary_id is None:
            return f"Tag '{self.name}' ({self.id})"
        else:
            return f"Tag '{self.name}' ({self.id}) [vocabulary {self.vocabulary_id}]"

    def to_dict(self, include_details:bool=True) -> dict:
        d = dict()
        if self.details is not None and include_details:
            d.update(self.details)
        d.update({"id": self.id, "name": self.name, "display_name": self.display_name,
                  "vocabulary_id": self.vocabulary_id})
        if self.state is not None:
            d["state"] = self.state
        return d

    @staticmethod
    def from_dict(d:dict) -> "CkanTagInfo":
        return CkanTagInfo(d)


class CkanResourceInfo(CkanConfigurableObjectABC):
    mandatory_attributes = {"name"}
    configurable_attributes = {"name", "state", "format", "description"}
    extra_attributes = {"download_url"}

    def __init__(self, d:dict=None, name:str=None, format:str=None, description:str=None, state:CkanState=None):
        self.id:Union[str,None] = None
        self.name:Union[str,None] = name
        self.package_id:Union[str,None] = None
        self.state:Union[CkanState,None] = state
        self.datastore_active:Union[bool,None] = None
        self.download_url:Union[str,None] = None
        self.format:Union[str,None] = format
        self.description:Union[str,None] = description
        self.datastore_info:Union[CkanDataStoreInfo,None] = None
        self.datastore_info_error:Union[dict,None] = None
        self.views:Union[OrderedDict[str,CkanViewInfo],None] = None  # dict id -> view info (list of known views - full list not guaranteed)
        self.view_is_full_list:bool = False
        self.created: Union[datetime.datetime,None] = None
        self.last_modified: Union[datetime.datetime,None] = None
        self.metadata_modified: Union[datetime.datetime,None] = None
        self.download_size_mb:Union[None,float] = None  # obtained through a HEAD request
        if d is not None:
            self.id = d["id"]
            self.name = d["name"]
            self.package_id = d["package_id"]
            if "state" in d.keys():
                self.state = CkanState.from_str(d["state"])
            self.datastore_active = d["datastore_active"]
            self.download_url = d["url"]
            self.format = d["format"]
            self.description = d["description"]
            if "datastore_info" in d.keys():
                self.datastore_info = CkanDataStoreInfo.from_dict(d["datastore_info"])
            if "datastore_info_error" in d.keys():
                self.datastore_info_error = d["datastore_info_error"]
            if "views" in d.keys():
                self.views = OrderedDict()
                for view_dict in d["views"]:
                    self.views[view_dict["id"]] = CkanViewInfo.from_dict(view_dict)
            self.created = datetime.datetime.fromisoformat(d["created"]) if "created" in d.keys() else None
            self.last_modified = datetime.datetime.fromisoformat(d["last_modified"]) if "last_modified" in d.keys() and d["last_modified"] is not None else None
            self.metadata_modified = datetime.datetime.fromisoformat(d["metadata_modified"]) if "metadata_modified" in d.keys() else None
        self.details:dict = d
        self.index_in_package:Union[int,None] = -1
        self.newly_created:bool = False
        self.newly_updated:bool = False

    def __str__(self):
        if self.datastore_info is not None:
            datastore_str = f"DataStore info"
        elif self.datastore_active:
            datastore_str = f"DataStore active"
        else:
            datastore_str = f"no DataStore"
        return f"Resource '{self.name}' ({self.id}) [{self.state}, {datastore_str}]"

    @staticmethod
    def get_resource_type() -> str:
        return "Resource"

    def copy(self) -> "CkanResourceInfo":
        return copy.deepcopy(self)

    def datastore_queried(self) -> bool:
        return self.datastore_info is not None or self.datastore_info_error is not None

    def update_view(self, view_info: Union[CkanViewInfo, List[CkanViewInfo]], view_list:bool=False) -> None:
        if isinstance(view_info, CkanViewInfo):
            view_info = [view_info]
        if self.views is None:
            self.views = OrderedDict()
        for view_info_update in view_info:
            self.views[view_info_update.id] = view_info_update
        self.view_is_full_list = self.view_is_full_list or view_list  # bool indicating if the list comes from full view list API

    def update(self, refresh: "CkanResourceInfo") -> None:
        self.id = refresh.id
        self.name = refresh.name
        self.state = refresh.state
        self.format = refresh.format
        self.description = refresh.description
        self.package_id = refresh.package_id
        self.datastore_active = refresh.datastore_active
        self.details = refresh.details

    def update_missing(self, refresh: "CkanResourceInfo") -> None:
        if self.id is None:
            self.id = refresh.id
        if self.name is None:
            self.name = refresh.name
        if self.state is None:
            self.state = refresh.state
        if self.format is None:
            self.format = refresh.format
        if self.description is None:
            self.description = refresh.description
        if self.details is not None:
            if refresh.details is not None:
                self.details.update(refresh.details)
        else:
            self.details = refresh.details

    def to_dict(self, include_details:bool=True) -> dict:
        d = dict()
        if self.details is not None and include_details:
            d.update(self.details)
        d.update({"id": self.id, "name": self.name, "package_id": self.package_id,
                 "datastore_active": self.datastore_active, "url": self.download_url,
                 "format": self.format, "description": self.description,
                  })
        if self.state is not None:
            d["state"] = str(self.state)
        if self.datastore_info is not None:
            d["datastore_info"] = self.datastore_info.to_dict(include_details=include_details)
        if self.datastore_info_error is not None:
            d["datastore_info_error"] = self.datastore_info_error
        if self.views is not None:
            d["views"] = [view.to_dict(include_details=include_details) for view in self.views.values()]
        if self.created is not None:
            d["created"] = self.created.isoformat()
        if self.last_modified is not None:
            d["last_modified"] = self.last_modified.isoformat()
        if self.metadata_modified is not None:
            d["metadata_modified"] = self.metadata_modified.isoformat()
        return d

    @staticmethod
    def from_dict(d:dict) -> "CkanResourceInfo":
        return CkanResourceInfo(d)


class CkanCollaboration:
    def __init__(self, capacity:CkanCapacity=None, modified:datetime.datetime=None, group_id:str=None, d:dict=None):
        self.capacity:CkanCapacity = capacity
        self.group_id:Union[str,None] = group_id
        self.modified: Union[datetime.datetime,None] = modified
        if d is not None:
            self.capacity = CkanCapacity.from_str(d["capacity"])
            self.modified:datetime.datetime = datetime.datetime.fromisoformat(d["modified"])

    def __str__(self):
        return str(self.capacity)

    def copy(self) -> "CkanCollaboration":
        return copy.deepcopy(self)

    def to_dict(self, user_info: CkanUserInfo, group_table: Dict[str,CkanGroupInfo], date_format:str) -> dict:
        d = OrderedDict([
            ("full_name", user_info.fullname),
            ("capacity", str(self.capacity)),
        ])
        if user_info.organizations is not None:
            d["organizations"] = sorted(user_info.organizations)
        if self.modified is not None:
            if date_format is None:
                d["date_modified"] = self.modified.isoformat()
            else:
                d["date_modified"] = self.modified.strftime(date_format)
        if self.group_id is not None:
            d["from_group"] = group_table[self.group_id].name
        return d


class CkanPackageInfo(CkanConfigurableObjectABC):
    mandatory_attributes = {"name"}
    configurable_attributes = {"name", "state", "title", "description", "private", "version",
                             "author", "author_email", "maintainer", "maintainer_email"}
    extra_attributes = {"tags", "custom_fields"}

    def __init__(self, d:dict=None, *, package_name:str=None, package_id:str=None,
                 title:str=None, description:str=None, private:bool=None, state:CkanState=None, version:str=None,
                 url:str=None, tags:List[str]=None):
        self.id:Union[str, None] = package_id
        self.name:Union[str, None] = package_name
        self.title:Union[str, None] = title
        self.description:Union[str, None] = description
        self.private:Union[bool, None] = private
        self.state:Union[CkanState, None] = state
        self.version:Union[str, None] = version
        self.custom_fields:OrderedDict[str,str] = OrderedDict()  # key, value pairs
        self.details:dict = {}
        self.package_resources:OrderedDict[str,CkanResourceInfo] = OrderedDict()  # resource id -> info
        self.resources_id_index:Dict[str,str] = {}  # resource name -> id
        self.resources_id_index_counts:Dict[str,int] = {}  # resource name -> counter
        self.organization_info: Union[CkanOrganizationInfo, None] = None
        self.groups:List[CkanGroupInfo] = []
        self.license_id:Union[str, None] = None
        self.author:Union[str, None] = None
        self.author_email:Union[str, None] = None
        self.maintainer:Union[str, None] = None
        self.maintainer_email:Union[str, None] = None
        self.url:Union[str, None] = url
        self.tags:Union[List[str],None] = tags
        self.tags_info:Union[Dict[str, CkanTagInfo],None] = None  # dict tag name -> tag info
        self.metadata_created: Union[datetime.datetime,None] = None
        self.metadata_modified: Union[datetime.datetime,None] = None
        self.requested_datastore_info:bool = False
        self.newly_created:bool = False
        self.collaborators:Union[None,Dict[str,CkanCollaboration]] = None  # given by API package_collaborator_list
        self.user_access:Union[None,Dict[str,CkanCollaboration]] = None  # given by function map_user_rights

        if d is not None:
            self.id = d["id"]
            self.name = d["name"]
            self.title = d["title"]
            self.description = d["notes"]
            self.private = d["private"]
            if "state" in d.keys():
                self.state = CkanState.from_str(d["state"])
            self.version = d["version"]
            self.custom_fields = OrderedDict()
            for field in d["extras"]:
                self.custom_fields[field["key"]] = field["value"]
            self.details = d
            self.package_resources = OrderedDict()
            for resource_info_dict in d["resources"]:
                self.package_resources[resource_info_dict["id"]] = CkanResourceInfo(resource_info_dict)
            self.resources_id_index = {resource_info.name: resource_info.id for resource_info in self.package_resources.values()}  # resource name -> id
            self.resources_id_index_counts = {}  # resource name -> counter
            for resource_info in self.package_resources.values():
                if resource_info.name not in self.resources_id_index_counts.keys():
                    self.resources_id_index_counts[resource_info.name] = 1
                else:
                    self.resources_id_index_counts[resource_info.name] += 1
            self.organization_info = None
            if "organization" in d.keys():
                if d["organization"] is not None:
                    self.organization_info = CkanOrganizationInfo(d["organization"])
                    assert_or_raise(self.organization_info.id == d["owner_org"], IntegrityError("Unexpected: organization != owner_org"))
                else:
                    assert_or_raise(d["owner_org"] is None, IntegrityError("Unexpected: organization != owner_org"))
            else:
                assert_or_raise("owner_org" not in d.keys() or d["owner_org"] == "", IntegrityError("Unexpected: organization is not present but owner_org was found"))
            self.groups = [CkanGroupInfo(info) for info in d["groups"]]
            self.license_id = d["license_id"]
            self.author = d["author"]
            self.author_email = d["author_email"]
            self.maintainer = d["maintainer"]
            self.maintainer_email = d["maintainer_email"]
            self.metadata_created = datetime.datetime.fromisoformat(d["metadata_created"]) if "metadata_created" in d.keys() else None
            self.metadata_modified = datetime.datetime.fromisoformat(d["metadata_modified"]) if "metadata_modified" in d.keys() else None
            self.url = d["url"]
            self.tags_info = {tag_dict["name"]: CkanTagInfo(tag_dict) for tag_dict in d["tags"]} if d["tags"] is not None else None
            self.tags = list(self.tags_info.keys()) if self.tags_info is not None else None
            self.updated:bool = False

    def __str__(self):
        return f"Package '{self.name}' ({self.id}) [{self.state}]"

    @staticmethod
    def get_resource_type() -> str:
        return "Package"

    def copy(self) -> "CkanPackageInfo":
        return copy.deepcopy(self)

    def update(self, refresh: "CkanPackageInfo"):
        refresh: CkanPackageInfo
        self.id = refresh.id
        self.name = refresh.name
        self.state = refresh.state
        self.details = refresh.details
        self.package_resources = refresh.package_resources
        self.resources_id_index = refresh.resources_id_index
        self.resources_id_index_counts = refresh.resources_id_index_counts

    def get_resource_index(self, resource_id:str) -> int:
        i_found = None
        for i, res_info in enumerate(self.package_resources.values()):
            if res_info.id == resource_id:
                i_found = i
                break
        return i_found

    def update_resource(self, resource_info: CkanResourceInfo) -> int:
        if resource_info.id is None:
            raise MissingIdError("Resource", resource_info.name)
        resource_id = resource_info.id
        i_update = self.get_resource_index(resource_id)
        if i_update is not None:
            resource_info.index_in_package = i_update
            self.package_resources[resource_id] = resource_info
        else:
            i_update = len(self.package_resources) - 1
            resource_info.index_in_package = i_update
            self.package_resources[resource_id] = resource_info
            self.resources_id_index_counts[resource_info.name] = 1
        self.resources_id_index[resource_info.name] = resource_id
        return i_update

    def to_dict(self, include_details:bool=True) -> dict:
        d = dict()
        if self.details is not None and include_details:
            d.update(self.details)
        d.update({"id": self.id, "name": self.name, "title": self.title,
                 "notes": self.description, "private": self.private, "version": self.version,
                 "tags": self.tags, "url": self.url,
                 "author": self.author, "author_email": self.author_email,
                 "maintainer": self.maintainer, "maintainer_email": self.maintainer_email,
                 "groups": [group.to_dict(include_details=include_details) for group in self.groups],
                 "license_id": self.license_id,
                 "resources": [resource.to_dict(include_details=include_details) for resource in self.package_resources.values()],
                 })
        if self.metadata_created is not None:
            d["metadata_created"] = self.metadata_created.isoformat()
        if self.metadata_modified is not None:
            d["metadata_modified"] = self.metadata_modified.isoformat()
        if self.state is not None:
            d["state"] = str(self.state)
        if self.custom_fields is not None:
            d["extras"] = [{"key": key, "value": value} for key, value in self.custom_fields.items()]
        if self.organization_info is not None:
            d["owner_org"] = self.organization_info.id
            d["organization"] = self.organization_info.to_dict(include_details=include_details)
        return d

    @staticmethod
    def from_dict(d:dict) -> "CkanPackageInfo":
        return CkanPackageInfo(d)


class CkanOrganizationInfo:
    def __init__(self, d:dict):
        self.id = d["id"]
        self.name = d["name"]
        self.state = CkanState.from_str(d["state"])
        self.title = d["title"]
        self.user_members: Union[None,Dict[str,CkanCapacity]] = None
        self.details = d
        if "users" in d:
            self.user_members = {user_dict["id"]: CkanCapacity.from_str(user_dict["capacity"]) for user_dict in d["users"]}

    def __str__(self):
        return f"Organization '{self.name}' ({self.id}) [{self.state}]"

    def copy(self) -> "CkanOrganizationInfo":
        return copy.deepcopy(self)

    def get_owner_org(self):
        """
        Returns the value used for the owner_org argument

        :return:
        """
        return self.name

    def to_dict(self, include_details:bool=True) -> dict:
        d = dict()
        if self.details is not None and include_details:
            d.update(self.details)
        d.update({"id": self.id, "name": self.name, "title": self.title, "state": str(self.state)})
        return d

    @staticmethod
    def from_dict(d:dict) -> "CkanOrganizationInfo":
        return CkanOrganizationInfo(d)




#!python3
# -*- coding: utf-8 -*-
"""
Code to initiate a DataStore defined by a large number of files to concatenate into one table
"""
import threading
from threading import Semaphore
from abc import ABC
from typing import Dict, List, Any, Tuple, Generator, Union, Set, Collection
from warnings import warn
from collections import OrderedDict

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_progress_callbacks import CkanCallbackLevel
from ckanapi_harvesters.builder.builder_resource_datastore import BuilderDataStoreABC
from ckanapi_harvesters.auxiliary.ckan_model import UpsertChoice
from ckanapi_harvesters.auxiliary.ckan_configuration import datastore_default_upload_index_col_name, datastore_default_source_file_col_name
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.builder.mapper_datastore_multi import RequestFileMapperABC, default_file_mapper_from_primary_key
from ckanapi_harvesters.builder.builder_resource_multi_abc import FileChunkDataFrame
from ckanapi_harvesters.builder.builder_resource_multi_file import BuilderMultiABC
from ckanapi_harvesters.builder.builder_field import BuilderField

# apply last_condition for each upsert request when in a multi-threaded upload on a same DataStore:
datastore_multi_threaded_always_last_condition:bool = True
# when there are multiple files, apply last insertion commands after each document? True: after each csv file, False: only at the end
datastore_multi_apply_last_condition_intermediary:bool = False


class BuilderDataStoreMultiABC(BuilderDataStoreABC, BuilderMultiABC, ABC):
    """
    generic class to manage large DataStore, divided into files/parts
    This abstract class is intended to be overloaded in order to be used to generate data from the workspace, without using CSV files
    """

    def __init__(self, *, parent, name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None, options_string:str=None, base_dir:str=None):
        super().__init__(parent=parent, name=name, format=format, description=description, resource_id=resource_id,
                         download_url=download_url, options_string=options_string, base_dir=base_dir)
        # Functions inputs/outputs
        self.df_mapper: RequestFileMapperABC
        self.setup_default_file_mapper()
        self.reupload_if_needed = False  # do not reupload if needed because this could cause data loss (the upload function only uploads the first table)
        self.upsert_method: UpsertChoice = UpsertChoice.Upsert
        # BuilderMultiABC:
        self.stop_event = threading.Event()
        self.thread_ckan: Dict[str, CkanApi] = {}
        self.enable_multi_threaded_upload:bool = True
        self.enable_multi_threaded_download:bool = True
        self.file_semaphore = Semaphore()
        self.process_level:CkanCallbackLevel = CkanCallbackLevel.ResourceChunks  # resource builder - can be changed to 3 for a file of a multi-file resource

    def copy(self, *, dest=None, parent=None):
        super().copy(dest=dest, parent=parent)
        dest.reupload_if_needed = self.reupload_if_needed
        # BuilderMultiABC:
        dest.progress_callback = self.progress_callback.copy()
        dest.enable_multi_threaded_upload = self.enable_multi_threaded_upload
        dest.enable_multi_threaded_download = self.enable_multi_threaded_download
        # do not copy stop_event
        return dest

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None) -> None:
        super()._load_from_df_row(row=row, base_dir=base_dir)
        self.setup_default_file_mapper()

    def _update_metadata(self, ckan: CkanApi, *, base_dir:str=None) -> None:
        """
        In certain implementations, the resource & field metadata can be derived from the data source.
        Normally, the metadata is defined by the user in an Excel worksheet. When a description is left empty,
        the value left on the CKAN server is left unchanged.
        The objective here is to propose values that override the Excel worksheet when the description
        is empty on the CKAN side (still leave CKAN values unchanged, if present).

        :param ckan: CkanApi instance
        :param override_ckan: when True, override the values from the CKAN server, if present
        """
        super()._update_metadata(ckan=ckan, base_dir=base_dir)
        if self.primary_key is not None and len(self.primary_key) == 1 and self.primary_key[0] == datastore_default_upload_index_col_name:
            if self.field_builders_user is None:
                self.field_builders_user = OrderedDict()
            if datastore_default_upload_index_col_name not in self.field_builders_user.keys():
                self.field_builders_user[datastore_default_upload_index_col_name] = BuilderField(name=datastore_default_upload_index_col_name, type_override="int8")
            field_builder = self.field_builders_user[datastore_default_upload_index_col_name]
            known_field = None
            if self.known_resource_info is not None and self.known_resource_info.datastore_info is not None and self.known_resource_info.datastore_info.fields_dict is not None:
                if datastore_default_upload_index_col_name in self.known_resource_info.datastore_info.fields_dict.keys():
                    known_field = self.known_resource_info.datastore_info.fields_dict[datastore_default_upload_index_col_name]
            if known_field is None or known_field.notes is None:
                field_builder.description = "Index of the line in the upload process"

    def setup_default_file_mapper(self, *, primary_key:List[str]=None, file_query_list:Collection[Tuple[str, dict]]=None) -> None:
        """
        This function enables the user to define the primary key and initializes the default file mapper.

        :param primary_key: manually specify the primary key
        :return:
        """
        df_mapper_mem = self.df_mapper
        if primary_key is None:
            if self.primary_key_user is not None:
                self.primary_key = self.primary_key_user
        else:
            self.primary_key = primary_key
        if (self.primary_key is None or len(self.primary_key) == 0) and self.column_enable_upload_index:
            self.primary_key = [datastore_default_upload_index_col_name]
        self.df_mapper = default_file_mapper_from_primary_key(self.primary_key, file_query_list)
        if self.column_enable_source_file:
            self.df_mapper.source_file_column = datastore_default_source_file_col_name
        if file_query_list is not None:
            self.downloaded_file_query_list = file_query_list
        # preserve upload/download functions
        self.df_mapper.df_upload_fun = df_mapper_mem.df_upload_fun
        self.df_mapper.df_download_fun = df_mapper_mem.df_download_fun

    ## upload ---------
    # do not change default argument apply_last_condition=True
    # def upsert_request_df(self, ckan: CkanApi, df_upload:pd.DataFrame,
    #                       method:UpsertChoice=UpsertChoice.Upsert,
    #                       apply_last_condition:bool=None) -> Tuple[pd.DataFrame, pd.DataFrame]:
    #     # calls super method, with apply_last_condition defaulting to datastore_multi_apply_last_condition_intermediary
    #     if apply_last_condition is None:
    #         apply_last_condition = True  # datastore_multi_apply_last_condition_intermediary
    #     return super().upsert_request_df(ckan=ckan, df_upload=df_upload, method=method,
    #                                      apply_last_condition=apply_last_condition)

    def _get_primary_key_indexes(self, data_cleaner_index: Set[str], current_df_fields:Set[str], error_missing:bool, empty_datastore:bool=False) -> Tuple[Union[List[str],None], Union[List[str],None]]:
        primary_key, indexes = super()._get_primary_key_indexes(data_cleaner_index, current_df_fields, error_missing, empty_datastore)
        # it is highly recommended to specify a primary key: warning if not defined
        if primary_key is None:
            msg = f"It is highly recommended to specify the primary key for a DataStore defined from a directory to ensure no duplicate values are upserted to the database. Resource: {self.name}"
            warn(msg)
        else:
            ultra_required_fields = set(primary_key)
            missing_fields = ultra_required_fields
            if current_df_fields is not None:
                missing_fields -= current_df_fields
            if len(missing_fields) > 0:
                msg = f"The primary key {self.primary_key} is set for resource {self.name} but it is not present in the sample data."
                warn(msg)
        if primary_key is None or len(primary_key) == 0:
            self.upsert_method = UpsertChoice.Insert  # do not use upsert
        return primary_key, indexes

    def upsert_request_final(self, ckan: CkanApi, *, force:bool=False) -> None:
        """
        Final steps after the last upsert query.
        This call is mandatory at the end of all requests if the user called upsert_request_df for a multi-part DataStore manually.

        :param ckan:
        :param force: perform request anyways
        :return:
        """
        force = force or not datastore_multi_apply_last_condition_intermediary
        return super().upsert_request_final(ckan, force=force)

    def upload_request_final(self, ckan: CkanApi, *, force:bool=False) -> None:
        super().upload_request_final(ckan, force=force)
        return self.upsert_request_final(ckan=ckan, force=force)

    def upsert_request_df_no_return(self, ckan: CkanApi, df_upload:pd.DataFrame, *,
                                    total_lines_read:int, file_name:str,
                                    method:UpsertChoice=UpsertChoice.Upsert,
                                    apply_last_condition:bool=None, always_last_condition:bool=None) -> None:
        """
        Calls upsert_request_df but does not return anything

        :return:
        """
        self.upsert_request_df(ckan=ckan, df_upload=df_upload, total_lines_read=total_lines_read, file_name=file_name,
                               method=method,
                               apply_last_condition=apply_last_condition, always_last_condition=always_last_condition)
        return None

    def _unit_upload_apply(self, *, ckan: CkanApi, file_chunk: FileChunkDataFrame,
                           upload_alter:bool=True, overall_chunk_index: int, file_count: int, start_index: int, end_index: int,
                           method: UpsertChoice, **kwargs) -> None:
        file_index = file_chunk.file_index
        if file_index == 0 and file_chunk.chunk_index == 0 and self.upsert_method == UpsertChoice.Insert:
            return  # do not reupload the first document, which was used for the initialization of the dataset
        process_upsert = start_index <= file_index and file_index < end_index and file_chunk.read_line_counter - len(file_chunk.df) >= self.upload_start_line
        self.progress_callback.task_progress(self.get_local_file_offset(file_chunk), self.get_local_file_total_size(),
                                             info=file_chunk, level=self.process_level,
                                             file_index=file_index, file_count=file_count,
                                             lines_chunk=len(file_chunk.df), total_lines_read=file_chunk.read_line_counter,
                                             canceled_request=not process_upsert,
                                             context=f"{ckan.identifier} upload")
        if process_upsert:
            # if upload_alter:
            #     file_chunk.df = self.df_mapper.df_upload_alter(file_chunk.df, self.sample_data_source, fields=self._get_fields_info(), **kwargs)
            # alter function is applied in upsert_request_df
            self.upsert_request_df_no_return(ckan=ckan, df_upload=file_chunk.df, method=method,
                                             total_lines_read=file_chunk.read_line_counter, file_name=file_chunk.file_path,
                                             apply_last_condition=datastore_multi_apply_last_condition_intermediary)
        else:
            # self._call_progress_callback(index, total, info=df_upload_local, context=f"{ckan.identifier} single-thread skip")
            pass

    def get_datastore_len(self, ckan:CkanApi) -> int:
        datastore_info = ckan.get_datastore_info_or_request(self.name, self.package_name, error_not_found=True)
        return datastore_info.row_count

    def upload_request_full(self, ckan:CkanApi, resources_base_dir:str, *,
                            method:UpsertChoice=None,
                            threads:int=1, external_stop_event=None,
                            allow_chunks:bool=True,
                            only_missing:bool=False, from_line_count:bool=False,
                            start_index:int=0, end_index:int=None, **kwargs) -> None:
        self.df_mapper.upsert_only_missing_rows = only_missing
        if from_line_count:
            self.upload_start_line = self.get_datastore_len(ckan)
            if self.upload_start_line is None:
                self.upload_start_line = 0
            if ckan.params.verbose_extra:
                print(f"Transfer will start from line {self.upload_start_line}")
        if method is None:
            if self.primary_key is None or len(self.primary_key) == 0:
                self.upsert_method = UpsertChoice.Insert  # do not use upsert if there is no primary key
            method = self.upsert_method
        super().upload_request_full(ckan=ckan, resources_base_dir=resources_base_dir,
                                    threads=threads, external_stop_event=external_stop_event,
                                    allow_chunks=allow_chunks,
                                    start_index=start_index, end_index=end_index,
                                    method=method, **kwargs)
        # if threads < 0:
        #     # cancel large uploads in this case
        #     return None
        # elif threads is None or threads > 1:
        #     return self.upload_request_full_multi_threaded(resources_base_dir=resources_base_dir, ckan=ckan, method=method,
        #                                                    threads=threads, external_stop_event=external_stop_event,
        #                                                    start_index=start_index, end_index=end_index)
        # else:
        #     self.init_local_files_list(resources_base_dir=resources_base_dir, cancel_if_present=True)
        #     if ckan.verbose_extra:
        #         print(f"Launching single-threaded upload of multi-file resource {self.name}")
        #     total = self.get_local_file_len()
        #     end_index = positive_end_index(end_index, total)
        #     for index, file in enumerate(self.get_local_file_generator(resources_base_dir=resources_base_dir)):
        #         if external_stop_event is not None and external_stop_event.is_set():
        #             print(f"{ckan.identifier} Interrupted")
        #             return
        #         self._unit_upload_apply(ckan=ckan, file=file,
        #                                 index=index, start_index=start_index, end_index=end_index, total=total,
        #                                 method=method)
        #     self._call_progress_callback(total, total, context=f"{ckan.identifier} single-thread upload")
        #     # at last, apply final actions:
        #     self.upload_request_final(ckan, force=not datastore_multi_apply_last_condition_intermediary)

    # def upsert_request_file_graceful(self, ckan: CkanApi, file: Any, index:int,
    #                                  method: UpsertChoice = UpsertChoice.Upsert, external_stop_event=None,
    #                                  start_index:int=0, end_index:int=None) -> None:
    #     """
    #     Calls upsert_request_df_clear with checks specific to multi-threading.
    #
    #     :return:
    #     """
    #     # ckan.session_reset()
    #     # ckan.identifier = current_thread().name
    #     ckan = self.thread_ckan[current_thread().name]
    #     total = self.get_local_file_len()
    #     end_index = positive_end_index(end_index, total)
    #     if self.stop_event.is_set():
    #         return
    #     if external_stop_event is not None and external_stop_event.is_set():
    #         print(f"{ckan.identifier} Interrupted")
    #         return
    #     try:
    #         self._unit_upload_apply(ckan=ckan, file=file,
    #                                 index=index, start_index=start_index, end_index=end_index, total=total,
    #                                 method=method)
    #     except Exception as e:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print(f"Stopping all threads because an exception occurred in thread: {e}")
    #         raise e from e

    # def upload_request_full_multi_threaded(self, ckan: CkanApi, resources_base_dir: str, threads: int = None,
    #                                        method: UpsertChoice = UpsertChoice.Upsert, external_stop_event=None,
    #                                        start_index:int=0, end_index:int=None, **kwargs):
    #     """
    #     Multi-threaded implementation of upload_request_full, using ThreadPoolExecutor.
    #     """
    #     self.init_local_files_list(resources_base_dir=resources_base_dir, cancel_if_present=True)
    #     resource_id = self.get_or_query_resource_id(ckan=ckan, error_not_found=True)  # prepare CKAN object for multi-threading: perform mapping requests if necessary
    #     self._prepare_for_multithreading(ckan)
    #     try:
    #         with ThreadPoolExecutor(max_workers=threads, initializer=self._init_thread, initargs=(ckan,)) as executor:
    #             if ckan.verbose_extra:
    #                 print(f"Launching multi-threaded upload of multi-file resource {self.name}")
    #             futures = [executor.submit(self.upsert_request_file_graceful, ckan=ckan, file=file, method=method, index=index,
    #                                        start_index=start_index, end_index=end_index, external_stop_event=external_stop_event)
    #                        for index, file in enumerate(self.get_local_file_generator(resources_base_dir=resources_base_dir))]
    #             for future in futures:
    #                 future.result()  # This will propagate the exception
    #         total = self.get_local_file_len()
    #         self._call_progress_callback(total, total, context=f"{ckan.identifier} multi-thread upload")
    #     except Exception as e:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print(f"Stopping all threads because an exception occurred: {e}")
    #         raise e from e
    #     finally:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print("End of multi-threaded upload...")
    #     # at last, apply final actions:
    #     self.upload_request_final(ckan, force=not datastore_multi_apply_last_condition_intermediary)


    ## download -------
    def download_file_query_generator(self, ckan: CkanApi, file_query:dict) -> Generator[pd.DataFrame, Any, None]:
        """
        Download the DataFrame with the file_query arguments
        """
        resource_id = self.get_or_query_resource_id(ckan, error_not_found=self.download_error_not_found)
        if resource_id is None and not self.download_error_not_found:
            return None
        download_generator = self.df_mapper.download_file_query(ckan=ckan, resource_id=resource_id, file_query=file_query,
                                                                progress_callback=self.progress_callback)
        for df_download in download_generator:
            df = self.df_mapper.df_download_alter(df_download, file_query=file_query, fields=self._get_fields_info())
            yield df

    def _unit_download_apply(self, ckan:CkanApi, file_query_item:Any, out_dir:str,
                            index:int, start_index:int, end_index:int, total:int,
                            **kwargs) -> Any:
        if start_index <= index and index < end_index:
            self.progress_callback.task_progress(index, total, info=file_query_item, level=self.process_level,
                                                 file_index=index, file_count=total,
                                                 total_lines_read=self.read_line_counter,
                                                 context=f"{ckan.identifier} single-thread download")
            self.download_file_query_item(ckan=ckan, out_dir=out_dir, file_query_item=file_query_item)
        else:
            pass
            # self._call_progress_callback(index, total, info=file_query_item, context=f"{ckan.identifier} single-thread skip")

    def download_request_full(self, ckan: CkanApi, out_dir: str, threads:int=1, external_stop_event=None,
                              start_index:int=0, end_index:int=None, force:bool=False) -> None:
        return super().download_request_full(ckan=ckan, out_dir=out_dir,
                                             threads=threads, external_stop_event=external_stop_event,
                                             start_index=start_index, end_index=end_index, force=force)
        # if (not self.enable_download) and (not force):
        #     msg = f"Did not download resource {self.name} because download was disabled."
        #     warn(msg)
        #     return None
        # if threads < 0:
        #     # do not download large datasets in this case
        #     return None
        # elif threads is None or threads > 1:
        #     return self.download_request_full_multi_threaded(ckan=ckan, out_dir=out_dir,
        #                                                      threads=threads, external_stop_event=external_stop_event,
        #                                                      start_index=start_index, end_index=end_index)
        # else:
        #     self.init_download_file_query_list(ckan=ckan, out_dir=out_dir, cancel_if_present=True)
        #     if ckan.verbose_extra:
        #         print(f"Launching single-threaded download of multi-file resource {self.name}")
        #     total = self.get_file_query_len()
        #     end_index = positive_end_index(end_index, total)
        #     for index, file_query_item in enumerate(self.get_file_query_generator()):
        #         if external_stop_event is not None and external_stop_event.is_set():
        #             print(f"{ckan.identifier} Interrupted")
        #             return
        #         self._unit_download_apply(ckan=ckan, file_query_item=file_query_item,
        #                                   index=index, start_index=start_index, end_index=end_index, total=total)
        #     self._call_progress_callback(total, total, context=f"{ckan.identifier} single-thread download")

    def download_request_generator(self, ckan: CkanApi, out_dir: str) -> Generator[Tuple[Any, pd.DataFrame], Any, None]:
        """
        Iterator on file_queries.
        """
        self.init_download_file_query_list(ckan=ckan, out_dir=out_dir, cancel_if_present=True)
        for file_query_item in self.get_file_query_generator():
            yield self.download_file_query_item(ckan=ckan, out_dir=out_dir, file_query_item=file_query_item)

    # def download_file_query_item_graceful(self, ckan: CkanApi, out_dir: str, file_query_item: Any, index:int,
    #                                       external_stop_event=None, start_index:int=0, end_index:int=None) -> None:
    #     """
    #     Implementation of download_file_query_item with checks for a multi-threaded download.
    #     """
    #     # ckan.session_reset()
    #     # ckan.identifier = current_thread().name
    #     ckan = self.thread_ckan[current_thread().name]
    #     total = self.get_file_query_len()
    #     end_index = positive_end_index(end_index, total)
    #     if self.stop_event.is_set():
    #         return
    #     if external_stop_event is not None and external_stop_event.is_set():
    #         print(f"{ckan.identifier} Interrupted")
    #         return
    #     try:
    #         # self._unit_download_apply(ckan=ckan, file_query_item=file_query_item,
    #         #                           index=index, start_index=start_index, end_index=end_index, total=total)
    #     except Exception as e:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print(f"Stopping all threads because an exception occurred in thread: {e}")
    #         raise e from e

    # def download_request_full_multi_threaded(self, ckan: CkanApi, out_dir: str,
    #                                          threads: int = None, external_stop_event=None,
    #                                          start_index:int=0, end_index:int=-1) -> None:
    #     """
    #     Multi-threaded implementation of download_request_full using ThreadPoolExecutor.
    #     """
    #     self.init_download_file_query_list(ckan=ckan, out_dir=out_dir, cancel_if_present=True)
    #     self._prepare_for_multithreading(ckan)
    #     try:
    #         with ThreadPoolExecutor(max_workers=threads, initializer=self._init_thread, initargs=(ckan,)) as executor:
    #             if ckan.verbose_extra:
    #                 print(f"Launching multi-threaded download of multi-file resource {self.name}")
    #             futures = [executor.submit(self.download_file_query_item_graceful, ckan=ckan, out_dir=out_dir, file_query_item=file_query_item,
    #                                        index=index, external_stop_event=external_stop_event, start_index=start_index, end_index=end_index)
    #                        for index, file_query_item in enumerate(self.get_file_query_generator())]
    #             for future in futures:
    #                 future.result()  # This will propagate the exception
    #         total = self.get_file_query_len()
    #         self._call_progress_callback(total, total, context=f"multi-thread download")
    #     except Exception as e:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print(f"Stopping all threads because an exception occurred: {e}")
    #         raise e from e
    #     finally:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print("End of multi-threaded download...")

    def download_resource_df(self, ckan: CkanApi, search_all:bool=False, **kwargs) -> pd.DataFrame:
        # alias with search_all=False by default
        return super().download_resource_df(ckan=ckan, search_all=search_all, **kwargs)

    def download_resource_bytes(self, ckan:CkanApi, full_download:bool=False, **kwargs) -> bytes:
        # alias with full_download=False by default
        return super().download_resource_bytes(ckan=ckan, full_download=full_download, **kwargs)





#!python3
# -*- coding: utf-8 -*-
"""
Extensions of os.path and operations on urls
"""
from typing import Union, Set, List
from warnings import warn
import os
import re


disable_relative_path_constraint = False

def unlock_relative_path_constraint(value:bool=True) -> None:
    """
    This function disables relative path error messages when a relative path is required.

    :return:
    """
    global disable_relative_path_constraint
    disable_relative_path_constraint = value
    if value:
        msg = "Relative path constraint is disabled to your own risk."
        warn(msg)


class BaseDirUndefError(Exception):
    def __init__(self, path: str):
        super().__init__(f"Could not determine the file path because no base_dir was provided: {path}")

class AbsolutePathError(Exception):
    def __init__(self, field:str, path: str):
        super().__init__(f"A relative path is highly suggested for this field ({field}): {path}. To disable this error message, run unlock_relative_path_constraint().")


def sanitize_path(path:Union[str,None],
                  *, expand_path:bool=False, keyword_exceptions:Set[str]=None) -> Union[str,None]:
    """
    Sanitize paths from user inputs
    """
    if path is not None:
        if os.path.sep == '\\':
            path = re.sub(r"[\\/]", "\\\\", path)
        else:
            path = re.sub(r"[\\/]", os.path.sep, path)
        if expand_path:
            path = os.path.expandvars(path)
            path_keyword = path.lower().strip()
            if path_keyword in keyword_exceptions:
                return path_keyword
            else:
                path = os.path.expanduser(path)  # cover the case where the path starts with '~'
        return path
    else:
        return None


def path_rel_to_dir(path:Union[str,None], base_dir:str=None, *, keyword_exceptions:Set[str]=None,
                    error_base_dir_undef:bool=False, default_value:str=None,
                    only_relative:bool=False, abs_error:bool=False, field:str=None) -> Union[str,None]:
    """
    Returns the absolute path. If relative, the base directory can be specified. If not specified, the cwd is used.

    :param path: original path string
    :param base_dir: the base directory, for relative paths if provided (default = cwd)
    :param keyword_exceptions: some values are not replaced and must be treated after this function call.
    :param error_base_dir_undef: Option to raise an error if no base_dir was provided (cwd is used by default).
    :param default_value: the value to return if path is None.
    :param only_relative: If set to True, a warning or error message is raised if an absolute path is provided.
    :param abs_error: Condition to choose between a warning or an error message.
    :param field: name of the field for the error message.
    :return: absolute path or keyword
    """
    if keyword_exceptions is None:
        keyword_exceptions = set()
    path_src = path
    path = sanitize_path(path)
    if path is None:
        return default_value
    else:
        path = os.path.expandvars(path)  # replace environment variables mentioned in the path, if any
        path_keyword = path.lower().strip()
        if path_keyword in keyword_exceptions:
            return path_keyword
        else:
            path = os.path.expanduser(path)  # cover the case where the path starts with '~'
            if os.path.isabs(path):
                if only_relative:
                    msg = AbsolutePathError(field, path_src)
                    if abs_error:
                        raise msg
                    else:
                        warn(str(msg))
                return path.strip()
            elif base_dir is not None:
                return os.path.join(base_dir, path.strip())
            elif error_base_dir_undef:
                raise BaseDirUndefError(path)
            else:
                return os.path.abspath(path)


def resolve_rel_path(base_dir:str, rel_path:str, *args: str, field:str, only_relative:bool=True) -> str:
    """
    Alias to path_rel_to_dir, with arguments order similar to os.path.join and requirement for a relative path.
    Relative path verification can be removed by calling unlock_relative_path_constraint.
    field: name of the field for the error message.

    :return:
    """
    if len(args) > 0:
        rel_path = os.path.join(rel_path, *args)
    return path_rel_to_dir(rel_path, base_dir=base_dir, field=field,
                           only_relative=only_relative, abs_error=not disable_relative_path_constraint,
                           error_base_dir_undef=True)


# reverse function:
def make_path_relative(path:str, to_base_dir:str = None, *, default_value:str=None,
                       source_string:str=None, keyword_exceptions:Set[str]=None, same_destination:bool=True) -> str:
    """
    When you save a file to a new location, make relative paths relative to the new file location,
    pointing to the same destination (except if same_destination is False -> source_string is used in this case, if present and relative path)
    The source_string is the path present in the original document.

    :param path: full file path (absolute, ideally output from path_rel_to_dir)
    :param to_base_dir: the new base directory, to derive the relative paths from
    :param default_value: the value to return if the path is None
    :param source_string: string representing the path in the original document, without any treatments
    :param keyword_exceptions: keywords to return as-is
    :return: path relative to to_base_dir or keyword/path relative to environment variable/home directory symbol (~)
    """
    if path is None:
        return default_value
    if source_string is not None:
        if (not same_destination) and (not os.path.isabs(source_string)):
            return source_string
        if keyword_exceptions is None:
            keyword_exceptions = set()
        source_string = sanitize_path(source_string)
        source_keyword = source_string.lower().strip()
        if source_keyword in keyword_exceptions:
            return source_keyword
        elif source_string.startswith('~') or source_string.startswith('$'):
            return source_string
        # elif not (os.path.expanduser(source_string) == source_string or os.path.expandvars(source_string) == source_string):
        #     # condition to confirm
        #     return source_string
    if path.startswith('~') or path.startswith('$'):
        return path
    else:
        return path if to_base_dir is None else os.path.relpath(path, to_base_dir)

# File search
def list_files_scandir(path:str) -> List[str]:
    # see also: glob.glob - this does not apply any filter
    list_files = []
    with os.scandir(path) as entries:
        for entry in entries:
            if entry.is_file():
                list_files.append(entry.path)
            elif entry.is_dir():
                list_files += list_files_scandir(entry.path)
        entries.close()
    return list_files

glob_chars = r"*?![]"
glob_re = "[\\*\\?\\!\\[\\]]+"

def glob_rm_glob(glob_str:str, *, default_rec_dir:str=None) -> str:
    """
    Extract directory name from a glob string (first elements of path without glob characters).

    :param glob_str: the glob string
    :param default_rec_dir: if the last removed element is "**" (directory recursion), the name of the directory to use instead
    :return: a path without glob characters

    Examples:
    >>> glob_rm_glob("test\\*.csv")
    'test'

    >>> glob_rm_glob("**\\*.csv", default_rec_dir="hello")
    'hello'
    """
    glob_free = glob_str
    while re.search(glob_re, glob_free):
        glob_free, glob_sub = os.path.split(glob_free)
        if glob_sub == "**" and re.search(glob_re, glob_free) is None and default_rec_dir is not None:
            return os.path.join(glob_free, default_rec_dir)
    return glob_free

def glob_name(glob_str:str):
    """
    Extract file name glob from a glob string (last element of path, except if it is "**")

    :param glob_str:
    :return:

    Example:
    >>> glob_name("**\\*.csv")
    '*.csv'
    """
    glob_dir, glob_file = os.path.split(glob_str)
    if not glob_file == "**":
        return glob_file
    else:
        return ""




#!python3
# -*- coding: utf-8 -*-
"""
Code to initialize a resource builder from a row
"""
from typing import Union
from warnings import warn

import pandas as pd

from ckanapi_harvesters.ckan_api import  CkanApiMap
from ckanapi_harvesters.auxiliary.ckan_model import CkanResourceInfo, CkanDataStoreInfo
from ckanapi_harvesters.auxiliary.ckan_auxiliary import assert_or_raise, _string_from_element
from ckanapi_harvesters.auxiliary.ckan_defs import ckan_tags_sep
from ckanapi_harvesters.auxiliary.ckan_errors import (UnexpectedError)
from ckanapi_harvesters.builder.builder_errors import MissingDataStoreInfoError
from ckanapi_harvesters.builder.builder_resource import (BuilderResourceABC, BuilderFileBinary, BuilderUrl,
                                                         BuilderResourceUnmanaged)
from ckanapi_harvesters.builder.builder_resource_multi_file import BuilderMultiFile
from ckanapi_harvesters.builder.builder_resource_datastore import BuilderResourceIgnored
from ckanapi_harvesters.builder.builder_resource_datastore_multi_folder import BuilderDataStoreFolder
from ckanapi_harvesters.builder.builder_resource_datastore_multi_ckan import BuilderDataStoreCkan
from ckanapi_harvesters.builder.builder_resource_datastore_file import BuilderDataStoreFile
from ckanapi_harvesters.builder.builder_resource_multi_datastore import BuilderMultiDataStore
from ckanapi_harvesters.builder.builder_resource_datastore_url import BuilderDataStoreUrl
from ckanapi_harvesters.builder.builder_resource_datastore_multi_harvester import BuilderDataStoreHarvester
from ckanapi_harvesters.builder.builder_resource_datastore_unmanaged import BuilderDataStoreUnmanaged
from ckanapi_harvesters.builder.builder_field import BuilderField


import_as_folder_row_count_threshold: Union[int,None] = None


def init_resource_from_df(row: pd.Series, parent, base_dir:str=None) -> Union[BuilderResourceABC,None]:
    """
    Function mapping keywords to a resource builder type.

    :param row:
    :return:
    """
    mode = _string_from_element(row["mode"], empty_value="")
    mode = mode.lower().strip()
    if mode == "file":
        resource_builder = BuilderFileBinary(parent=parent)
    elif mode == "url":
        resource_builder = BuilderUrl(parent=parent)
    elif mode == "datastore from file":
        resource_builder = BuilderDataStoreFile(parent=parent)
    elif mode == "datastore from folder":
        resource_builder = BuilderDataStoreFolder(parent=parent)
    elif mode == "datastore from url":
        resource_builder = BuilderDataStoreUrl(parent=parent)
    elif mode == "datastore from harvester":
        resource_builder = BuilderDataStoreHarvester(parent=parent)
    elif mode == "unmanaged":
        resource_builder = BuilderResourceUnmanaged(parent=parent)
    elif mode == "unmanaged datastore":
        resource_builder = BuilderDataStoreUnmanaged(parent=parent)
    elif mode == "multifile":
        resource_builder = BuilderMultiFile(parent=parent)
    elif mode == "ckan datastore merge":
        resource_builder = BuilderDataStoreCkan(parent=parent)
    elif mode == "multidatastore":
        resource_builder = BuilderMultiDataStore(parent=parent)
    elif mode == "ignored":
        resource_builder = BuilderResourceIgnored(parent=parent)
    elif mode == "":
        resource_name = _string_from_element(row["name"], empty_value="", strip=True)
        if resource_name == "":
            msg = "Ignoring line with empty resource name and empty resource mode: " + row.to_json()
            warn(msg)
            return None
        else:
            raise ValueError(f"Empty resource mode for resource '{resource_name}'")
    else:
        raise ValueError(f"{mode} is not a valid mode")
    resource_builder._load_from_df_row(row=row, base_dir=base_dir)
    return resource_builder


def init_resource_from_ckan(ckan: CkanApiMap, resource_info: CkanResourceInfo, parent) -> BuilderResourceABC:
    """
    Function initiating a resource builder based on information provided by the CKAN API.

    :return:
    """
    # assert_or_raise(ckan.map._mapping_query_datastore_info, MissingDataStoreInfoError())
    assert_or_raise(resource_info.datastore_queried(), MissingDataStoreInfoError())
    d = {
        "name": resource_info.name,
        "format": resource_info.format,
        "description": resource_info.description,
        "state": resource_info.state.name if resource_info.state is not None else "",
        "file/url": resource_info.name,
        "primary key": "",
        "indexes": "",
        "known id": resource_info.id,
        "known url": resource_info.download_url,
    }
    if (isinstance(resource_info.datastore_info, CkanDataStoreInfo)
            and resource_info.datastore_info.row_count is not None
            and len(resource_info.datastore_info.fields_id_list) > 0):
        # DataStore
        d["indexes"] = ckan_tags_sep.join(resource_info.datastore_info.index_fields)
        d["aliases"] = ckan_tags_sep.join(resource_info.datastore_info.aliases)
        if len(resource_info.download_url) > 0 and not ckan.is_url_internal(resource_info.download_url):
            d["file/url"] = resource_info.download_url
            row = pd.Series(d)
            resource = BuilderDataStoreUrl(parent=parent)
            resource._load_from_df_row(row=row)
        elif resource_info.format.lower() == "csv":
            row = pd.Series(d)
            resource = BuilderDataStoreUnmanaged(parent=parent)
            resource._load_from_df_row(row=row)
            if import_as_folder_row_count_threshold is not None and resource_info.datastore_info.row_count > import_as_folder_row_count_threshold:
                resource = resource.to_builder_datastore_folder()
        else:
            raise UnexpectedError(f"Format of data store {resource_info.name} ({resource_info.format}) is not recognized")
        # load fields information
        resource.field_builders = {}
        for field_id in resource_info.datastore_info.fields_id_list:
            field_info = resource_info.datastore_info.fields_dict[field_id]
            resource.field_builders[field_id] = BuilderField._from_ckan_field(field_info)
    elif len(resource_info.download_url) > 0 and not ckan.is_url_internal(resource_info.download_url):
        # external resource
        d["file/url"] = resource_info.download_url
        row = pd.Series(d)
        resource = BuilderUrl(parent=parent)
        resource._load_from_df_row(row=row)
        assert_or_raise(not resource_info.datastore_active and not isinstance(resource_info.datastore_info, CkanResourceInfo), UnexpectedError())
    else:
        # file
        row = pd.Series(d)
        resource = BuilderResourceUnmanaged(parent=parent)
        resource._load_from_df_row(row=row)
    resource.package_name = resource_info.package_id
    return resource




#!python3
# -*- coding: utf-8 -*-
"""
Data format policy representation and enforcing
"""
from typing import Tuple, Union
from enum import IntEnum
from collections import OrderedDict


newline_char = '\n'


class ListChoiceMode(IntEnum):
    Any = 0
    MaxOne = 1
    MandatoryOne = 2
    MandatoryMulti = 3
    NoExtra = 4

    def __str__(self):
        return self.name.lower()

    @staticmethod
    def from_str(s):
        s = s.lower().strip()
        if s == "any":
            return ListChoiceMode.Any
        elif s == "maxone":
            return ListChoiceMode.MaxOne
        elif s == "mandatoryone":
            return ListChoiceMode.MandatoryOne
        elif s == "mandatorymulti":
            return ListChoiceMode.MandatoryMulti
        elif s == "noextra":
            return ListChoiceMode.NoExtra
        else:
            raise ValueError(s)


class StringMatchMode(IntEnum):
    Any = 0
    NotEmpty = 1
    Match = 2
    MatchCaseSensitive = 3
    Regex = 4
    RegexCaseSensitive = 5
    Wildcard = 6
    WildcardCaseSensitive = 7

    def __str__(self):
        return self.name.lower()

    @staticmethod
    def from_str(s):
        s = s.lower().strip()
        if s == "any":
            return StringMatchMode.Any
        elif s == "notempty":
            return StringMatchMode.NotEmpty
        elif s == "match":
            return StringMatchMode.Match
        elif s == "matchcasesensitive":
            return StringMatchMode.MatchCaseSensitive
        elif s == "regex":
            return StringMatchMode.Regex
        elif s == "regexcasesensitive":
            return StringMatchMode.RegexCaseSensitive
        elif s == "wildcard":
            return StringMatchMode.Wildcard
        elif s == "wildcardcasesensitive":
            return StringMatchMode.WildcardCaseSensitive
        else:
            return StringMatchMode.Any  # default value


class DataType(IntEnum):
    Text = 1
    Numeric = 2
    TimeStamp = 3
    Bool = 4

    def __str__(self):
        return self.name.lower()

    @staticmethod
    def from_str(s):
        s = s.lower().strip()
        if s == "text":
            return DataType.Text
        elif s == "numeric":
            return DataType.Numeric
        elif s == "timestamp":
            return DataType.TimeStamp
        elif s == "bool":
            return DataType.Bool
        else:
            raise ValueError(s)


class StringValueSpecification:
    def __init__(self, value:str, help:str=None):
        if help is None:
            help = ""
        self.value: str = value
        self.help: str = help

    def to_tuple(self) -> Tuple[str,str]:
        return self.value, self.help

    @staticmethod
    def from_tuple(values: Union[str, Tuple[str,str]]) -> "StringValueSpecification":
        if isinstance(values, str):
            value, help = values, ""
        elif len(values) == 1:
            value, help = values[0], ""
        else:
            value, help = values
        return StringValueSpecification(value, help)

    def to_dict(self) -> dict:
        return OrderedDict([("value", self.value), ("help", self.help)])

    @staticmethod
    def from_dict(values: dict) -> "StringValueSpecification":
        value = values["value"]
        help = values["help"] if "help" in values.keys() else ""
        return StringValueSpecification(value, help)



